%{
/* syntax analyzer for casm ver.1.0 ['92/Aug/25] */

#include <stdio.h>
#include <malloc.h>
#include "defines.h"
#include "const.h"
#include "opcode.h"

/* max instruction word length */
#define MAXINSTLEN 5

/* bit mask pattern */
static int bitmask[]={  0,
			0x1,       0x3,       0x7,       0xf,
			0x1f,      0x3f,      0x7f,      0xff,
			0x1ff,     0x3ff,     0x7ff,     0xfff,
			0x1fff,    0x3fff,    0x7fff,    0xffff,
			0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
			0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
			0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
			0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

/* enable uninitialized label used error flag */
int	uninitlabelerr = FALSE;

/* instruction struct */
typedef struct {
	int	optype;
	int	subtype;
	int	src1flag;
	int	src1;
	int	src2flag;
	int	src2;
	int	dstnflag;
	int	dstn;
} insttype;

insttype inst;

/* composed instruction word */
int	instword[MAXINSTLEN];

/* PC increment value */
int	incpc;

/* DW pseudo instruction counter */
int	dwcount;

/* check bit width */
int	checkwidth();

/* error print routine */
void	yyerror();

/* error flag */
int	errorflag = FALSE;

%}
%union {
	int	intvalue;
	char	*charptr;
}
%left OROP
%left XOROP
%left ANDOP
%left RSFTOP LSFTOP
%left ADDOP SUBOP
%left MULOP DIVOP MODOP
%right UMINUS UPLUS NOTOP
%token LP RP
%token LD LDI ST STI PUSH POP PUSHI MOV
%token ADD ADDI SUB SUBI MUL MULI
%token DIV DIVI MOD MODI RSUB RSUBI
%token AND ANDI OR ORI XOR XORI
%token NEG NOT NEGI NOTI
%token ORG EQU DW DS
%token NOP HLT
%token COLON COMMA IMMED
%token <intvalue> REG INTVAL
%token <charptr> LABEL NL
%type <intvalue> intexp pseudoinst source1 source2 dstn
%start program
%%
program
	: linelist
	;

linelist
	: /* empty */
	| linelist line
	;

line
	: NL
		{
		if (pass == 2) {
		    listoutcmnt($1);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$1);
		    errorflag = FALSE;
		}
		}
	| LABEL COLON NL
		{
		if (pass == 1) {
		    addlabelandinit($1,currentpc);
		} else { /* pass 2 */
		    listoutwithpc(currentpc,$3);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$3);
		    errorflag = FALSE;
		}
		}
	| LABEL COLON instruction NL
		{
		int	i;

		if (pass == 1) {
		    addlabelandinit($1,currentpc);
		} else { /* pass 2 */
		    for (i = 0; i < incpc; i += 3) {
			listoutwithinst(currentpc + i,
					((incpc - i)>=3) ? 3 : (incpc - i) % 3,
					instword + i,
					(i == 0) ? $4 : "\n");
		    } 
		    for (i = 0; i < incpc ; i++) {
			objectout(instword[i]);
		    }
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$4);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| instruction NL
		{
		int	i,j;

		if (pass == 2) {
		    for (i = 0; i < incpc; i += 3) {
			listoutwithinst(currentpc + i,
					((incpc - i)>=3) ? 3 : (incpc - i) % 3,
					instword + i,
					(i == 0) ? $2 : "\n");
		    } 
		    for (i = 0; i < incpc ; i++) {
			objectout(instword[i]);
		    }
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$2);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| LABEL COLON pseudoinst NL
		{
		int	i;
		if (pass == 1) {
		    addlabelandinit($1,currentpc);
		} else { /* pass 2 */
		    switch ($3) {
			case ORG:
			    listoutwithpc(currentpc,$4);
			    objectaddrout(currentpc);
			    break;
			case DW:
			    for (i = 0 ; i < dwcount ; i+=3) {
				listoutwithinst(currentpc + i,
						((dwcount - i)>=3) ? 3 : (dwcount - i) % 3,
						dwbuff + i,
						(i == 0) ? $4 : "\n");
			    }
			    for (i = 0 ; i < dwcount ; i++) {
				objectout(dwbuff[i]);
			    }
			    break;
			case DS:
			    instword[0] = 0;
			    listoutwithinst(currentpc,
					    1,
					    instword,
					    $4);
			    for (i = 0 ; i < incpc ; i++) {
				objectout(0);
			    }
			    break;
			default:
			    listoutwithpc(currentpc,$4);
			    yyerror("irregal pseudo instruction");
			    break;
		    }
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$4);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| pseudoinst NL
		{
		int	i;
		if (pass == 2) {
		    switch ($1) {
			case ORG:
			    listoutwithpc(currentpc,$2);
			    objectaddrout(currentpc);
			    break;
			case DW:
			    for (i = 0 ; i < dwcount ; i+=3) {
				listoutwithinst(currentpc + i,
						((dwcount - i)>=3) ? 3 : (dwcount - i) % 3,
						dwbuff + i,
						(i == 0) ? $2 : "\n");
			    }
			    for (i = 0 ; i < dwcount ; i++) {
				objectout(dwbuff[i]);
			    }
			    break;
			case DS:
			    instword[0] = 0;
			    listoutwithinst(currentpc,
					    1,
					    instword,
					    $2);
			    for (i = 0 ; i < incpc ; i++) {
				objectout(0);
			    }
			    break;
			default:
			    listoutcmnt($2);
			    yyerror("irregal pseudo instruction");
			    break;
		    }
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$2);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| LABEL EQU {uninitlabelerr = TRUE;} intexp NL
		{
		if (pass == 1) {
		    addlabelandinit($1,$4);
		} else { /* pass 2 */
		    listoutequ(getlabelvalue($1),$5);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$5);
		    errorflag = FALSE;
		}
		uninitlabelerr = uninitlabelerrdefault;
		}
	| error NL 
		{
		fprintf(stderr,$2);
		if (pass == 2) {
		    listoutcmnt($2);
		}
		errorflag = FALSE;
		yyerrok;
		}
	;

instruction
	: loadstore
	| intopinst
	| unaryinst
	| miscinst
	;

pseudoinst
	: ORG {uninitlabelerr = TRUE;} intexp
		{
		currentpc = $3;
		incpc = 0 ;
		uninitlabelerr = uninitlabelerrdefault;
		$$ = ORG;
		}
	| DW {dwcount = 0;} dwbody
		{
		if (pass == 1) {
			if (dwcount > maxdwc) maxdwc = dwcount;
			incpc = dwcount;
		} else { /* pass 2 */
			incpc = dwcount;
		}
		$$ = DW;
		}
	| DS {uninitlabelerr = TRUE;} intexp
		{
		incpc = $3 ;
		uninitlabelerr = uninitlabelerrdefault;
		$$ = DS;
		}
	;

dwbody
	: dwterm
	| dwbody COMMA dwterm
	;

dwterm
	: intexp
		{
		if (pass == 1) {
			dwcount++;
		} else {
			dwbuff[dwcount++] = $1;
		}
		}
	;

loadstore
	: LD intexp COMMA REG
		{
		if (pass == 2) {
		    if (checkwidth($2,16)!=TRUE) {
			yyerror("error : LD address out of range");
			$2 = 0;
		    }
		    instword[0] =
			(LOADSTORE << 13) |
			(LS_LOAD << 9) |
			(MEMOPR << 6) ;
		    instword[1] = (REGOPR << 6) | $4;
		    instword[2] = $2 ;
		}
		incpc = 3;
		}
	| LD IMMED intexp COMMA REG
		{
		if (pass == 2) {
		    if (checkwidth($3,16)!=TRUE) {
			yyerror("error : LD immediate value out of range");
			$3 = 0;
		    }
		    instword[0] =
			(LOADSTORE << 13) |
			(LS_LOAD << 9) |
			(IMMOPR << 6) ;
		    instword[1] = (REGOPR << 6) | $5;
		    instword[2] = $3 ;
		}
		incpc = 3;
		}
	| ST REG COMMA intexp
		{
		if (pass == 2) {
		    if (checkwidth($4,16)!=TRUE) {
			yyerror("error : ST address out of range");
			$4 = 0;
		    }
		    instword[0] =
			(LOADSTORE << 13) |
			(LS_STORE << 9) |
			(REGOPR << 6) | $2;
		    instword[1] = (MEMOPR << 6) ;
		    instword[2] = $4 ;
		}
		incpc = 3;
		}
	| ST IMMED intexp COMMA intexp
		{
		if (pass == 2) {
		    if (checkwidth($3,16)!=TRUE) {
			yyerror("error : ST immediate value out of range");
			$3 = 0;
		    }
		    if (checkwidth($5,16)!=TRUE) {
			yyerror("error : ST address out of range");
			$5 = 0;
		    }
		    instword[0] =
			(LOADSTORE << 13) |
			(LS_STORE << 9) |
			(IMMOPR << 6) ;
		    instword[1] = (MEMOPR << 6) ;
		    instword[2] = $3 ;
		    instword[3] = $5 ;
		}
		incpc = 4;
		}
	| MOV REG COMMA REG
		{
		if (pass == 2) {
		    instword[0] =
			(LOADSTORE << 13) |
			(LS_LOAD << 9) |
			(REGOPR << 6) | $2 ;
		    instword[1] = (REGOPR << 6) | $4;
		}
		incpc = 2;
		}
	| MOV intexp COMMA intexp
		{
		if (pass == 2) {
		    if (checkwidth($2,16)!=TRUE) {
			yyerror("error : MOV source address out of range");
			$2 = 0;
		    }
		    if (checkwidth($4,16)!=TRUE) {
			yyerror("error : MOV distination address out of range");
			$4 = 0;
		    }
		    instword[0] =
			(LOADSTORE << 13) |
			(LS_STORE << 9) |
			(MEMOPR << 6) ;
		    instword[1] = (MEMOPR << 6) ;
		    instword[2] = $2 ;
		    instword[3] = $4 ;
		}
		incpc = 4;
		}
	;

miscinst
	: NOP
		{
		instword[0] =
		    (MISC << 13) | 
		    (MISC_NOP << 9) ;
		incpc = 1;
		}	
	| HLT
		{
		instword[0] =
		    (MISC << 13) | 
		    (MISC_HLT << 9) ;
		incpc = 1;
		}
	;

source1
	: IMMED intexp
		{
		if (checkwidth($2,16)!=TRUE) {
		    yyerror("error : source1 immediate value out of range");
		    $2 = 0;
		}
		inst.src1flag = (IMMOPR << 6);
		inst.src1 = $2;
		$$ = IMMOPR;
		}
	| REG
		{
		inst.src1flag = (REGOPR << 6) | $1;
		$$ = REGOPR;
		}
	| intexp
		{
		if (checkwidth($1,16)!=TRUE) {
		    yyerror("error : source1 memory address out of range");
		    $1 = 0;
		}
		inst.src1flag = (MEMOPR << 6);
		inst.src1 = $1;
		$$ = MEMOPR;
		}
	;

source2
	: IMMED intexp
		{
		if (checkwidth($2,16)!=TRUE) {
		    yyerror("error : source2 immediate value out of range");
		    $2 = 0;
		}
		inst.src2flag = (IMMOPR << 6);
		inst.src2 = $2;
		$$ = IMMOPR;
		}
	| REG
		{
		inst.src2flag = (REGOPR << 6) | $1;
		$$ = REGOPR;
		}
	| intexp
		{
		if (checkwidth($1,16)!=TRUE) {
		    yyerror("error : source2 memory address out of range");
		    $1 = 0;
		}
		inst.src2flag = (MEMOPR << 6);
		inst.src2 = $1;
		$$ = MEMOPR;
		}
	;

dstn
	: REG
		{
		inst.dstnflag = (REGOPR << 6) | $1;
		$$ = REGOPR;
		}
	| intexp
		{
		if (checkwidth($1,16)!=TRUE) {
		    yyerror("error : distination memory address out of range");
		    $1 = 0;
		}
		inst.dstnflag = (MEMOPR << 6);
		inst.dstn = $1;
		$$ = MEMOPR;
		}
	;
	
unaryinst
	: NEG source1 COMMA dstn
		{
		instword[0] =
		    (INTCALC1 << 13) | 
		    (UOP_NEG << 9) |
		    inst.src1flag ;
		instword[1] = inst.dstnflag;
		switch ($2) {
		    case IMMOPR:
				if ($4==REGOPR) { /* NEG I,R */
				    instword[2] = inst.src1;
				    incpc = 3;
				} else {	  /* NEG I,M */
				    instword[2] = inst.src1;
				    instword[3] = inst.dstn;
				    incpc = 4;
				}
				break;
		    case MEMOPR:
				if ($4==REGOPR) { /* NEG M,R */
				    instword[2] = inst.src1;
				    incpc = 3;
				} else {	  /* NEG M,M */
				    instword[2] = inst.src1;
				    instword[3] = inst.dstn;
				    incpc = 4;
				}
				break;
		    case REGOPR:
				if ($4==REGOPR) { /* NEG R,R */
				    incpc = 2;
				} else {	  /* NEG R,M */
				    instword[2] = inst.dstn;
				    incpc = 3;
				}
				break;
		}
		}	
	| NOT source1 COMMA dstn
		{
		instword[0] =
		    (INTCALC1 << 13) | 
		    (UOP_NOT << 9) |
		    inst.src1flag ;
		instword[1] = inst.dstnflag;
		switch ($2) {
		    case IMMOPR:
				if ($4==REGOPR) { /* NEG I,R */
				    instword[2] = inst.src1;
				    incpc = 3;
				} else {	  /* NEG I,M */
				    instword[2] = inst.src1;
				    instword[3] = inst.dstn;
				    incpc = 4;
				}
				break;
		    case MEMOPR:
				if ($4==REGOPR) { /* NEG M,R */
				    instword[2] = inst.src1;
				    incpc = 3;
				} else {	  /* NEG M,M */
				    instword[2] = inst.src1;
				    instword[3] = inst.dstn;
				    incpc = 4;
				}
				break;
		    case REGOPR:
				if ($4==REGOPR) { /* NEG R,R */
				    incpc = 2;
				} else {	  /* NEG R,M */
				    instword[2] = inst.dstn;
				    incpc = 3;
				}
				break;
		}
		}
	;

intopinst
	: int_ops source1 COMMA source2 COMMA dstn
		{
		instword[0] =
		    (INTCALC2 << 13) |
		    (inst.subtype << 9) |
		    inst.src1flag ;
		instword[1] = 
		    (inst.src2flag << 8) |
		    inst.dstnflag ;
		if ($2 == REGOPR) { /* source1 = reg (op R,*,*) */
		    switch ($4) {
			case IMMOPR:
				    if ($6==REGOPR) { /* op R,I,R */
					instword[2] = inst.src2;
					incpc = 3;
				    } else {	  /* op R,I,M */
					instword[2] = inst.src2;
					instword[3] = inst.dstn;
					incpc = 4;
				    }
				    break;
			case MEMOPR:
				    if ($6==REGOPR) { /* op R,M,R */
					instword[2] = inst.src2;
					incpc = 3;
				    } else {	  /* op R,M,M */
					instword[2] = inst.src2;
					instword[3] = inst.dstn;
					incpc = 4;
				    }
				    break;
			case REGOPR:
				    if ($6==REGOPR) { /* op R,R,R */
					incpc = 2;
				    } else {	  /* op R,R,M */
					instword[2] = inst.dstn;
					incpc = 3;
				    }
				    break;
		    }
		} else {	/* source1 = imm/mem (op [IM],*,*) */
		    switch ($4) {
			case IMMOPR:
				    if ($6==REGOPR) { /* op [IM],I,R */
					instword[2] = inst.src1;
					instword[3] = inst.src2;
					incpc = 4;
				    } else {	  /* op [IM],I,M */
					instword[2] = inst.src1;
					instword[3] = inst.src2;
					instword[4] = inst.dstn;
					incpc = 5;
				    }
				    break;
			case MEMOPR:
				    if ($6==REGOPR) { /* op [IM],M,R */
					instword[2] = inst.src1;
					instword[3] = inst.src2;
					incpc = 4;
				    } else {	  /* op [IM],M,M */
					instword[2] = inst.src1;
					instword[3] = inst.src2;
					instword[4] = inst.dstn;
					incpc = 5;
				    }
				    break;
			case REGOPR:
				    if ($6==REGOPR) { /* op [IM],R,R */
					instword[2] = inst.src1;
					incpc = 3;
				    } else {	  /* op [IM],R,M */
					instword[2] = inst.src1;
					instword[3] = inst.dstn;
					incpc = 4;
				    }
				    break;
		    }
		}
		}
	;

int_ops
	: ADD
		{
		inst.subtype = OP_ADD;
		}
	| SUB
		{
		inst.subtype = OP_SUB;
		}
	| MUL
		{
		inst.subtype = OP_MUL;
		}
	| DIV
		{
		inst.subtype = OP_DIV;
		}
	| AND
		{
		inst.subtype = OP_AND;
		}
	| OR
		{
		inst.subtype = OP_OR;
		}
	| XOR
		{
		inst.subtype = OP_XOR;
		}
	| MOD
		{
		inst.subtype = OP_MOD;
		}
	;

intexp
	: LABEL
		{
		$$ = getlabelvalue($1);
		}
	| INTVAL
		{
		$$ = $1;
		}
	| LP intexp RP
		{
		$$ = $2;
		}
	| intexp ADDOP intexp
		{
		$$ = $1 + $3;
		}
	| intexp SUBOP intexp
		{
		$$ = $1 - $3;
		}
	| intexp MULOP intexp
		{
		$$ = $1 * $3;
		}
	| intexp DIVOP intexp
		{
		if ($3 != 0) 
		    $$ = $1 / $3;
		else {
		    if (uninitlabelerr == TRUE) {
			yyerror("divide by zero in integer expression");
		    }
		    $$ = 0;
		}
		}
	| intexp MODOP intexp
		{
		if ($3 != 0) 
			$$ = $1 % $3;
		else {
		    if (uninitlabelerr == TRUE) {
			yyerror("modulo by zero in integer expression");
		    }
		    $$ = 0;
		}
		}
	| intexp RSFTOP intexp
		{
		$$ = $1 >> $3;
		}
	| intexp LSFTOP intexp
		{
		$$ = $1 << $3;
		}
	| intexp ANDOP intexp
		{
		$$ = $1 & $3;
		}
	| intexp OROP intexp
		{
		$$ = $1 | $3;
		}
	| intexp XOROP intexp
		{
		$$ = $1 ^ $3;
		}
	| ADDOP intexp %prec UPLUS
		{
		$$ = $2;
		}
	| SUBOP intexp %prec UMINUS
		{
		$$ = -$2;
		}
	| NOTOP intexp
		{
		$$ = ~$2;
		}
	;
%%
/*
   search label , and return its pointer.
   if not found , add label list.
   (labelvalue is not initialized)
*/
labelnode *addlabel(name)
char *name;
{
	labelnode	*lnp,*lastlnp,*newlnp;
#ifdef DEBUG
	fprintf(stderr,"addlabel %s\n",name);
#endif

	if (root == NULL) { /* no label added */
		root = labelnodealloc();
		root->labelname = name;
		root->labelvalue = 0;
		root->initialized = FALSE;
		root->next = NULL;
		return root;
	} else { /* already some labels added */ 
		lnp = root;
		while (lnp != NULL) {
			if (strcmp(lnp->labelname,name) == 0) {
				return (lnp);
			}
			lastlnp = lnp;
			lnp = lnp->next;
		}
		if (lnp == NULL) { /* not found */
			newlnp = labelnodealloc();
			lastlnp->next = newlnp;
			newlnp->labelname = name;
			newlnp->labelvalue = 0;
			newlnp->initialized = FALSE;
			newlnp->next = NULL;
			return newlnp;
		}
	}
}

/*
   search label , set label value and return its pointer.
   if not found , add label tree and set label value.
*/
labelnode *addlabelandinit(name,value)
char	*name;
int	value;
{
	labelnode *foundlabel;
	char	errorstr[256];

#ifdef DEBUG
	fprintf(stderr,"addlabelandinit %s,%d\n",name,value);
#endif
	foundlabel = addlabel(name);

	foundlabel->labelvalue = value;
#ifdef DEBUG
	fprintf(stderr,"addlabelandinit label=%s value=(%d)\n",name,foundlabel->labelvalue);
#endif

	if (foundlabel->initialized == TRUE) {
		sprintf(errorstr,"warning: label redefined. (%s)\n",
			foundlabel->labelname);
		yyerror(errorstr);
	}
	foundlabel->initialized = TRUE;
	
	return foundlabel;
}

/*
   search label , and get it's value.
   if not found or not initialized, returns 0.
*/
int getlabelvalue(name)
char	*name;
{
	labelnode *foundlabel;
	int	returnvalue;
	char	errorstr[256];

#ifdef DEBUG
	fprintf(stderr,"getlabelvalue %s\n",name);
#endif
	foundlabel = addlabel(name);
#ifdef DEBUG
	fprintf(stderr,"getlabelvalue label=%s value=(%d)\n",name,foundlabel->labelvalue);
#endif

	if (foundlabel->initialized == TRUE) {
	    returnvalue = foundlabel->labelvalue;
	} else { /* labelvalue is not initialized */
	    if (uninitlabelerr == TRUE) {
		/* uninitlabelused = TRUE; */
		sprintf(errorstr,"warning: uninitialized label used. (%s)\n",
			foundlabel->labelname);
		yyerror(errorstr);
	    }
	    returnvalue = 0;
	}
	
	return returnvalue;
}

/*
   allocate memory for labelnode
*/
labelnode *labelnodealloc()
{
	return (labelnode *) malloc(sizeof(labelnode));
}

/* printout yacc error message */
void yyerror(s)
char *s;
{
	extern int yynerrs,yylineno;

	++yynerrs;
	fprintf(stderr,"%d: %s\n",yylineno,s);
	listoutcmnt(s);
	errorflag = TRUE;
}

/* check bit width (OK = return TRUE , else return FALSE) */
int	checkwidth(value,width)
int	value,width;
{
	int check;

	if (value < 0) {
		check = (value & bitmask[width])|(~bitmask[width]);
	} else {
		check = value & bitmask[width];
	}

	if (value != check) return FALSE;
	else return TRUE;
}
