#include <stdio.h>
#include "defines.h"

#define CODE_STACKLEN 50
#define NOOFREG 8

typedef struct {
	int	type;
	char 	var_name[128];
	int	value;
} CS;

static CS   code_stack[CODE_STACKLEN];
static int  code_stacktop = -1;

#ifndef SIMPLETMP
#define MAXTMP 50
static int called=FALSE;
static int reguse[NOOFREG];
static int tmpuse[MAXTMP];
#endif

code_push(cs)	/* stack push (for code-generation) */
CS	*cs;
{
#ifdef DEBUG
	printf ("code_push:push (");
	print_cs(*cs);
	printf(")\n");
#endif
	if (++code_stacktop == CODE_STACKLEN) { /* stack overflow */
		error("code stack overflow");
		return;
	} else {
		code_stack[code_stacktop].type = cs->type;
		switch (cs->type) {
		    case VARNAME:
			strcpy(code_stack[code_stacktop].var_name,
				   cs->var_name);
			break;
		    case CONST:
		    case TEMP:
		    case REG:
			code_stack[code_stacktop].value = cs->value;
			break;
		}
	}
}

CS *code_pop() /* stack pop (for code-generation) */
{
	if (code_stacktop == -1) { /* stack underflow */
		error("code stack underflow");
		return;
	} else {
#ifdef DEBUG
		printf("code_pop:pop (");
		print_cs(code_stack[code_stacktop]);
		printf(")\n");
#endif
		return &code_stack[code_stacktop--];
	}
}

print_cs(cs)
CS cs;
{
	switch (cs.type) {
	    case VARNAME:
	    	printf("type:VARNAME %s",cs.var_name);
	    	break;
	    case CONST:
	    	printf("type:CONST %d",cs.value);
	    	break;
	    case TEMP:
	    	printf("type:TEMP T%d",cs.value);
	    	break;
	    case REG:
	    	printf("type:REG R%d",cs.value);
	    	break;
	}
}

code_gen()
{
	REVP *rp;
	CS   cs,*left,*right;
	static char op[10],src1[256],src2[256],dst[256];

#ifdef DEBUG
	printf("generated code is:\n");
#endif

	rp = revp_top;
	while (rp != NULL) {
	    switch (rp->type) {
		case SYMBOL:
		    if (rp->symbol != '=') {
			right = code_pop();
			left = code_pop();
			switch (left->type) {
			    case VARNAME:
				sprintf(src1,"Var_%s",left->var_name);
				break;
			    case TEMP:
				sprintf(src1,"Tmp_T%d",left->value);
				temp_free(left);
				break;
			    case REG:
				sprintf(src1,"R%d",left->value);
				temp_free(left);
				break;
			    case CONST:
				sprintf(src1,"#%d",left->value);
				break;
			}
			switch (rp->symbol) {
			    case '+':
				strcpy (op,"ADD");
				break;
			    case '-':
				strcpy (op,"SUB");
				break;
			    case '*':
				strcpy (op,"MUL");
				break;
			    case '/':
				strcpy (op,"DIV");
				break;
			}
			switch (right->type) {
			    case VARNAME:
				sprintf(src2,"Var_%s",right->var_name);
				break;
			    case TEMP:
				sprintf(src2,"Tmp_T%d",right->value);
				temp_free(right);
				break;
			    case REG:
				sprintf(src2,"R%d",right->value);
				temp_free(right);
				break;
			    case CONST:
				sprintf(src2,"#%d",right->value);
				break;
			}
			get_temp(&cs);
			switch (cs.type) {
			    case TEMP:
				sprintf(dst,"Tmp_T%d",cs.value);
				break;
			    case REG:
				sprintf(dst,"R%d",cs.value);
				break;
			}
			printf("\t%s\t%s,%s,%s\n",op,src1,src2,dst);
			code_push(&cs);
		    } else { /* rp->symbol == '=' */
			right = code_pop();
			left = code_pop();
			if (left->type != VARNAME) {
			    error("assignment target is not variable");
			} else {
			    switch (right->type) {
			        case VARNAME:
				    printf("\tMOV\tVar_%s,Var_%s\n",
				    	    right->var_name,
				    	    left->var_name);
				    break;
				case TEMP:
				    printf("\tMOV\tTmp_T%d,Var_%s\n",
				    	    right->value,
				    	    left->var_name);
				    temp_free(right);
				    break;
				case REG:
				    printf("\tST\tR%d,Var_%s\n",
				    	    right->value,
				    	    left->var_name);
				    temp_free(right);
				    break;
				case CONST:
				    printf("\tST\t#%d,Var_%s\n",
				    	    right->value,
				    	    left->var_name);
				    break;
			    }
			}
			return;
		    }
		    break;
		case VARNAME:
		    cs.type = rp->type;
		    strcpy(cs.var_name,rp->var_name);
		    code_push(&cs);
		    break;
		case CONST:
		    cs.type = rp->type;
		    cs.value = rp->value;
		    code_push(&cs);
		    break;
	    }
	    rp = rp->next;
	}
}

#ifdef SIMPLETMP
get_temp(cs)
CS *cs;
{
	static int tc = 0;

	if (tc < NOOFREG) { 
	    cs->type = REG;
	    cs->value = tc++;
	} else {
	    cs->type = TEMP;
	    cs->value = (tc++)-NOOFREG+1;
	    tmp_add(cs->value);
    	}
}

temp_free(cs)
CS *cs;
{
}
#else
init_temp()
{
	int i;

	for (i = 0 ; i < NOOFREG ; i++) reguse[i] = FALSE;
	for (i = 1 ; i < MAXTMP ; i++) tmpuse[i] = FALSE;
}

get_temp(cs)
CS	*cs;
{
	int i;

	if (called == FALSE) {
	    init_temp();
	    called = TRUE;
	}

	for (i = 0 ; i < NOOFREG ; i++) {
	    if (reguse[i] == FALSE) {
		reguse[i] = TRUE;
		cs->type = REG;
		cs->value = i;
		return ;
	    }
	}
	for (i = 1 ; i < MAXTMP ; i++) {
	    if (tmpuse[i] == FALSE) {
		tmpuse[i] = TRUE;
		cs->type = TEMP;
		cs->value = i;
		tmp_add(cs->value);
		return ;
	    }
	}
	error("temp area overflow");
}

temp_free(cs)
CS	*cs;
{
	if (cs->type == REG) {
	    reguse[cs->value] = FALSE;
	} else {
	    tmpuse[cs->value] = FALSE;
	}
}
#endif
