#include <stdio.h>
#include "defines.h"

#define REVP_STACKLEN 256

static char revp_stack[REVP_STACKLEN];
static int  revp_stacktop = -1;

revp_push(c)	/* stack push (for rev-polish conversion) */
char	c;
{
#ifdef DEBUG
	fprintf (stderr,"revp_push:push %c\n",c);
#endif
	if (++revp_stacktop == REVP_STACKLEN) { /* stack overflow */
		error("revp stack overflow");
		return;
	} else {
		revp_stack[revp_stacktop] = c;
	}
}

char revp_pop()	/* stack pop (for rev-polish conversion) */
{
	if (revp_stacktop == -1) { /* stack underflow */
		error("revp stack underflow");
		return;
	} else {
#ifdef DEBUG
		fprintf (stderr,"revp_pop:pop %c\n",revp_stack[revp_stacktop]);
#endif
		return revp_stack[revp_stacktop--];
	}
}

append_revp(w)
WORD w;
{
	REVP *new;

	new = (REVP *)malloc(sizeof(REVP));

	new->type = w.type;
	switch (new->type) {
		case SYMBOL:
			new->symbol = w.symbol;
			break;
		case CONST:
			new->value = w.value;
			break;
		case VARNAME:
			strcpy(new->var_name,w.var_name);
			break;
	}
	if (revp_top == NULL) { /* revp_list is empty */
		revp_top = new;
		revp_tail = new;
		revp_top->next = NULL;
	} else {		/* revp_list is not empty */
		revp_tail->next = new;
		revp_tail = new;
		revp_tail->next = NULL;
	}
	return;
}

print_revp()
{
	REVP *rp;
	WORD w;

	printf(";rev-polish notation is:\n");

	rp = revp_top;
	while (rp != NULL) {
	    w.type = rp->type;
	    switch (w.type) {
		case SYMBOL:
		    w.symbol = rp->symbol;
		    break;
		case VARNAME:
		    strcpy(w.var_name,rp->var_name);
		    break;
		case CONST:
		    w.value = rp->value;
	    }
	    printf(";");
	    printword(w);
	    printf("\n");
	    rp = rp->next;
	}
}

printword(w)
WORD w;
{
	switch (w.type) {
	    case VARNAME:
		printf("type:VARNAME ; name:%s",w.var_name);
		break;
	    case CONST:
		printf("type:CONST ; value:%d",w.value);
		break;
	    case SYMBOL:
		switch (w.symbol) {
		    case '+':
		    case '-':
		    case '*':
		    case '/':
		    case '(':
		    case ')':
		    case '=':
			printf("type:SYMBOL ; symbol:%c",(char)w.symbol);
			break;
		    case '\n':
			printf("type:SYMBOL ; symbol:CR");
			break;
		    case EOF:
			printf("type:SYMBOL ; symbol:EOF");
			break;
		}
		break;
	}
}

/* compare priority (alpha>beta:1,alpha=beta:0,alpha<beta:-1)*/
int compare(alpha,beta)
char alpha,beta;
{
	switch (alpha) {
	    case '=':
		switch (beta) {
		    case '=':
			return 0;
		    case '+':
		    case '-':
		    case '*':
		    case '/':
			return -1;
		}
		break;
	    case '+':
	    case '-':
		switch (beta) {
		    case '=':
			return 1;
		    case '+':
		    case '-':
			return 0;
		    case '*':
		    case '/':
			return -1;
		}
		break;
	    case '*':
	    case '/':
		switch (beta) {
		    case '=':
		    case '+':
		    case '-':
			return 1;
		    case '*':
		    case '/':
			return 0;
		}
		break;
	}
}

int conv_revp() /* rev-polish conversion (if EOF return EOF) */
{
	WORD w,pw;
	char alpha,beta;
	int  flag;
	char revp_pop();

	revp_top = NULL;

	for (;;) {
	    getword(&w);
#ifdef DEBUG
	    printword(w); printf("\n");
#endif
	    switch (w.type) {
		case VARNAME:
		case CONST:
		    append_revp(w);
		    break;
		case SYMBOL:
		    alpha = w.symbol;
		    switch (alpha) {
			case '+':
			case '-':
			case '*':
			case '/':
			case '=':
			    flag = TRUE;
			    while (flag==TRUE) {
				if (revp_stacktop == -1) {
				    revp_push((char)alpha);
				    flag = FALSE;
				} else if (revp_stack[revp_stacktop] == '(') { 
				    revp_push((char)alpha);
				    flag = FALSE;
				} else {
				    beta = revp_stack[revp_stacktop];
				    switch(compare(alpha,beta)) {
					case -1:
					case 0:
					    pw.symbol = revp_pop();
					    pw.type = SYMBOL;
					    append_revp(pw);
					    flag = TRUE;
					    break;
					case 1:
					    revp_push((char)w.symbol);
					    flag = FALSE;
				    }
				}
			    }
			    break;
			case '(':
			    revp_push((char)w.symbol);
			    break;
			case '\n':
			case ')':
			    flag = TRUE;
			    while (flag==TRUE) {
				beta = revp_pop();
				switch(beta) {
				    case '(':
					flag = FALSE;
					break;
				    case '+':
				    case '-':
				    case '*':
				    case '/':
					pw.symbol = beta;
					pw.type = SYMBOL;
					append_revp(pw);
					flag = TRUE;
					break;
				    case '=':
					pw.symbol = beta;
					pw.type = SYMBOL;
					append_revp(pw);
					line++;
					return 0;
					break;
				}
			    }
			    break;
			case EOF:
			    return EOF;
			    break;
		    }
		    break;
	    	case ERROR:
	    	    error ("irregal character");
	    	    return;
	    }
	}
}
