#include <stdio.h>
#include "defines.h"

#define CODE_STACKLEN 50
#define NOOFREG 32

typedef struct {
	int	type;
	char 	var_name[128];
	int	value;
} CS;

static CS   code_stack[CODE_STACKLEN];
static int  code_stacktop = -1;

#ifndef SIMPLETMP
#define MAXTMP 50
static int called=FALSE;
static int reguse[NOOFREG];
static int tmpuse[MAXTMP];
#endif

code_push(cs)	/* stack push (for code-generation) */
CS	*cs;
{
#ifdef DEBUG
	printf ("code_push:push (");
	print_cs(*cs);
	printf(")\n");
#endif
	if (++code_stacktop == CODE_STACKLEN) { /* stack overflow */
		error("code stack overflow");
		return;
	} else {
		code_stack[code_stacktop].type = cs->type;
		switch (cs->type) {
		    case VARNAME:
			strcpy(code_stack[code_stacktop].var_name,
				   cs->var_name);
			break;
		    case CONST:
		    case TEMP:
		    case REG:
			code_stack[code_stacktop].value = cs->value;
			break;
		}
	}
}

CS *code_pop() /* stack pop (for code-generation) */
{
	if (code_stacktop == -1) { /* stack underflow */
		error("code stack underflow");
		return;
	} else {
#ifdef DEBUG
		printf("code_pop:pop (");
		print_cs(code_stack[code_stacktop]);
		printf(")\n");
#endif
		return &code_stack[code_stacktop--];
	}
}

print_cs(cs)
CS cs;
{
	switch (cs.type) {
	    case VARNAME:
	    	printf("type:VARNAME %s",cs.var_name);
	    	break;
	    case CONST:
	    	printf("type:CONST %d",cs.value);
	    	break;
	    case TEMP:
	    	printf("type:TEMP T%d",cs.value);
	    	break;
	    case REG:
	    	printf("type:REG R%d",cs.value);
	    	break;
	}
}

code_gen()
{
	REVP *rp;
	CS   cs,*left,*right;
	static char op[10],src1[256],src2[256],dst[256];

#ifdef DEBUG
	printf("generated code is:\n");
#endif

	rp = revp_top;
	while (rp != NULL) {
	    switch (rp->type) {
		case SYMBOL:
		    if (rp->symbol != '=') {
			switch (rp->symbol) {
			    case '+':
				strcpy (op,"ADD");
				break;
			    case '-':
				strcpy (op,"SUB");
				break;
			    case '*':
				strcpy (op,"MUL");
				break;
			    case '/':
				strcpy (op,"DIV");
				break;
			}
			right = code_pop();
			left = code_pop();
			switch (left->type) {
			    case VARNAME:
			    case TEMP:
				code_mxx(op,rp,left,right);
				break;
			    case REG:
				code_rxx(op,rp,left,right);
				break;
			    case CONST:
				code_ixx(op,rp,left,right);
				break;
			}
		    } else { /* rp->symbol == '=' */
			right = code_pop();
			left = code_pop();
			if (left->type != VARNAME) {
			    error("assignment target is not variable");
			} else {
			    switch (right->type) {
			        case VARNAME:
				    printf("\tLD\tVar_%s,R0\n",
				    	    right->var_name);
				    printf("\tST\tVar_%s,R0\n",
				    	    left->var_name);
				    break;
				case TEMP:
				    printf("\tLD\tTmp_T%d,R0\n",
				    	    right->value);
				    printf("\tST\tVar_%s,R0\n",
				    	    left->var_name);
				    temp_free(right);
				    break;
				case REG:
				    printf("\tST\tVar_%s,R%d\n",
				    	    left->var_name,
				    	    right->value);
				    temp_free(right);
				    break;
				case CONST:
				    printf("\tLDI\t%d,R0\n",
				    	    right->value);
				    printf("\tST\tVar_%s,R0\n",
				    	    left->var_name);
				    break;
			    }
			}
			return;
		    }
		    break;
		case VARNAME:
		    cs.type = rp->type;
		    strcpy(cs.var_name,rp->var_name);
		    code_push(&cs);
		    break;
		case CONST:
		    cs.type = rp->type;
		    cs.value = rp->value;
		    code_push(&cs);
		    break;
	    }
	    rp = rp->next;
	}
}

code_mxx(op,rp,left,right) /* memory op x -> x */
char *op;
REVP *rp;
CS   *left,*right;
{
	CS	cs;

	if (left->type == VARNAME) {
	    printf("\tLD\tVar_%s,R0\n",left->var_name);
	} else { /* left->type = TEMP */
	    printf("\tLD\tTmp_T%d,R0\n",left->value);
	    temp_free(left);
	}
	switch (right->type) {
	    case VARNAME: /* M op M -> x */
		printf("\tLD\tVar_%s,R1\n",right->var_name);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* M op M -> R */
			printf("\t%s\tR0,R1,R%d\n",op,cs.value);
			code_push(&cs);
			break;
		    case TEMP: /* M op M -> M */
			printf("\t%s\tR0,R1,R0\n",op);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			code_push(&cs);
			break;
		}
		break;
	    case TEMP: /* M op M -> x */
		printf("\tLD\tTmp_T%d,R1\n",right->value);
		temp_free(right);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* M op M -> R */
			printf("\t%s\tR0,R1,R%d\n",op,cs.value);
			break;
		    case TEMP: /* M op M -> M */
			printf("\t%s\tR0,R1,R0\n",op);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	    case REG: /* M op R -> x */
		temp_free(right);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* M op R -> R */
			printf("\t%s\tR0,R%d,R%d\n",op,right->value,cs.value);
			break;
		    case TEMP: /* M op R -> M */
			printf("\t%s\tR0,R%d,R0\n",op,right->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	    case CONST: /* M op I -> x */
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* M op I -> R */
			printf("\t%sI\tR0,%d,R%d\n",op,right->value,cs.value);
			break;
		    case TEMP: /* M op I -> M */
			printf("\t%sI\tR0,%d,R0\n",op,right->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	}
}

code_rxx(op,rp,left,right) /* register op x -> x */
char *op;
REVP *rp;
CS   *left,*right;
{
	CS   cs;

	temp_free(left);
	switch (right->type) {
	    case VARNAME: /* R op M -> x */
		printf("\tLD\tVar_%s,R1\n",right->var_name);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* R op M -> R */
			printf("\t%s\tR%d,R1,R%d\n",op,left->value,cs.value);
			code_push(&cs);
			break;
		    case TEMP: /* R op M -> M */
			printf("\t%s\tR%d,R1,R0\n",op,left->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			code_push(&cs);
			break;
		}
		break;
	    case TEMP: /* R op M -> x */
		printf("\tLD\tTmp_T%d,R1\n",right->value);
		temp_free(right);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* R op M -> R */
			printf("\t%s\tR%d,R1,R%d\n",op,left->value,cs.value);
			break;
		    case TEMP: /* R op M -> M */
			printf("\t%s\tR%d,R1,R0\n",op,left->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	    case REG: /* R op R -> x */
		temp_free(right);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* R op R -> R */
			printf("\t%s\tR%d,R%d,R%d\n",
				op,left->value,right->value,cs.value);
			break;
		    case TEMP: /* R op R -> M */
			printf("\t%s\tR%d,R%d,R0\n",
				op,left->value,right->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	    case CONST: /* R op I -> x */
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* R op I -> R */
			printf("\t%sI\tR%d,%d,R%d\n",
				op,left->value,right->value,cs.value);
			break;
		    case TEMP: /* R op I -> M */
			printf("\t%sI\tR%d,%d,R0\n",
				op,left->value,right->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	}
}

code_ixx(op,rp,left,right) /* immediate op x -> x */
char *op;
REVP *rp;
CS   *left,*right;
{
	CS   cs;

	printf("\tLDI\t%d,R0\n",left->value);
	switch (right->type) {
	    case VARNAME: /* I op M -> x */
		printf("\tLD\tVar_%s,R1\n",right->var_name);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* I op M -> R */
			printf("\t%s\tR0,R1,R%d\n",op,cs.value);
			code_push(&cs);
			break;
		    case TEMP: /* I op M -> M */
			printf("\t%s\tR0,R1,R0\n",op);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			code_push(&cs);
			break;
		}
		break;
	    case TEMP: /* I op M -> x */
		printf("\tLD\tTmp_T%d,R1\n",right->value);
		temp_free(right);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* I op M -> R */
			printf("\t%s\tR0,R1,R%d\n",op,cs.value);
			break;
		    case TEMP: /* I op M -> M */
			printf("\t%s\tR0,R1,R0\n",op);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	    case REG: /* I op R -> x */
		temp_free(right);
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* I op R -> R */
			printf("\t%s\tR0,R%d,R%d\n",op,right->value,cs.value);
			break;
		    case TEMP: /* I op R -> M */
			printf("\t%s\tR0,R%d,R0\n",op,right->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	    case CONST: /* I op I -> x */
		get_temp(&cs);
		switch (cs.type) {
		    case REG: /* I op I -> R */
			printf("\t%sI\tR0,%d,R%d\n",op,right->value,cs.value);
			break;
		    case TEMP: /* I op I -> M */
			printf("\t%sI\tR0,%d,R0\n",op,right->value);
			printf("\tST\tTmp_T%d,R0\n",cs.value);
			break;
		}
		code_push(&cs);
		break;
	}
}

#ifdef SIMPLETMP
get_temp(cs)
CS *cs;
{
	static int tc = 2;

	if (tc < NOOFREG) { 
	    cs->type = REG;
	    cs->value = tc++;
	} else {
	    cs->type = TEMP;
	    cs->value = (tc++)-NOOFREG+1;
	    tmp_add(cs->value);
    	}
}

temp_free(cs)
CS *cs;
{
}
#else
init_temp()
{
	int i;

	for (i = 2 ; i < NOOFREG ; i++) reguse[i] = FALSE;
	for (i = 1 ; i < MAXTMP ; i++) tmpuse[i] = FALSE;
}

get_temp(cs)
CS	*cs;
{
	int i;

	if (called == FALSE) {
	    init_temp();
	    called = TRUE;
	}

	for (i = 2 ; i < NOOFREG ; i++) {
	    if (reguse[i] == FALSE) {
		reguse[i] = TRUE;
		cs->type = REG;
		cs->value = i;
		return ;
	    }
	}
	for (i = 1 ; i < MAXTMP ; i++) {
	    if (tmpuse[i] == FALSE) {
		tmpuse[i] = TRUE;
		cs->type = TEMP;
		cs->value = i;
		tmp_add(cs->value);
		return ;
	    }
	}
	error("temp area overflow");
}

temp_free(cs)
CS	*cs;
{
	if (cs->type == REG) {
	    reguse[cs->value] = FALSE;
	} else {
	    tmpuse[cs->value] = FALSE;
	}
}
#endif
