#include <stdio.h>
#include "defines.h"

#define TREE_STACKLEN 256

static struct node *tree_s1[TREE_STACKLEN];
static int  tree_s1top = -1;
static char tree_s2[TREE_STACKLEN];
static int  tree_s2top = -1;

tree_stackinit()  /* tree conversion stack pointer initialize */
{
	tree_s1top = -1;
	tree_s2top = -1;
}

tree_s1push(node) /* stack S1 push (for tree conversion) */
struct node	*node;
{
#ifdef DEBUG
	printf ("tree_s1push:push node (");
	printword(node->w);
	printf (")\n");
#endif
	if (++tree_s1top == TREE_STACKLEN) { /* stack overflow */
		error("tree S1 stack overflow");
		return;
	} else {
		tree_s1[tree_s1top] = node;
	}
}

struct node *tree_s1pop() /* stack S1 pop (for tree conversion) */
{
	if (tree_s1top == -1) { /* stack underflow */
		error("tree S1 stack underflow");
		return;
	} else {
#ifdef DEBUG
		printf ("tree_s1pop:pop node (");
		printword(tree_s1[tree_s1top]->w);
		printf (")\n");
#endif
		return tree_s1[tree_s1top--];
	}
}

tree_s2push(c)	/* stack S2 push (for tree conversion) */
char	c;
{
#ifdef DEBUG
	printf ("tree_s2push:push %c\n",c);
#endif
	if (++tree_s2top == TREE_STACKLEN) { /* stack overflow */
		error("tree S2 stack overflow");
		return;
	} else {
		tree_s2[tree_s2top] = c;
	}
}

char tree_s2pop() /* stack S2 pop (for tree conversion) */
{
	if (tree_s2top == -1) { /* stack underflow */
		error("tree S2 stack underflow");
		return;
	} else {
#ifdef DEBUG
		printf ("tree_s2pop:pop %c\n",tree_s2[tree_s2top]);
#endif
		return tree_s2[tree_s2top--];
	}
}

struct node *makenode(word)
struct word word;
{
	struct node *new;

	if ((new = (struct node *)malloc(sizeof(struct node)))==NULL) {
	    error("node malloc fail");
	    exit(1);
	}

	new->w = word;
	new->right = NULL;
	new->left = NULL;
	return new;
}

print_tree()
{
	struct node *np;
	char s[256];

	printf(";tree notation is:\n");

	if (tree_root != NULL) {
	    s[0]='\0';
	    print_tree_sub(tree_root,1,s);
	} else {
	    printf("tree_root is NULL\n");
	}
}

print_tree_sub(node,level,s)
struct node *node;
int  level;
char *s;
{
	char tmp[256],ls[256],rs[256];

	if (node != NULL) {
	    sprintf(tmp,"%s%d",s,level);
	    printf(";level:%s : node = ",tmp);
	    printword(node->w);
	    printf("\n");
	    sprintf(ls,"%sL",tmp);
	    print_tree_sub(node->left,level+1,ls);
	    sprintf(rs,"%sR",tmp);
	    print_tree_sub(node->right,level+1,rs);
	} else {
	    /* printf("NULL\n"); */
	}
}

printword(w)
struct word w;
{
	switch (w.type) {
	    case VARNAME:
		printf("type:VARNAME ; name:%s",w.varname);
		break;
	    case CONST:
		printf("type:CONST ; value:%d",w.value);
		break;
	    case SYMBOL:
		switch (w.symbol) {
		    case '+':
		    case '-':
		    case '*':
		    case '/':
		    case '(':
		    case ')':
		    case '=':
			printf("type:SYMBOL ; symbol:%c",(char)w.symbol);
			break;
		    case '\n':
			printf("type:SYMBOL ; symbol:CR");
			break;
		    case EOF:
			printf("type:SYMBOL ; symbol:EOF");
			break;
		}
		break;
	    case REGNUM:
		printf("type:REGNUM ; value:%d",w.regnum);
		break;
	    case TMPNUM:
		printf("type:TMPNUM ; value:%d",w.tmpnum);
		break;
	}
}

/* yuusen jyun'i hikaku (alpha>beta:1, alpha=beta:0, alpha<beta:-1)*/
int compare(alpha,beta)
char alpha,beta;
{
#ifdef DEBUG
	printf("compare(%c,%c)\n",alpha,beta);
#endif

	switch (alpha) {
	    case '=':
		switch (beta) {
		    case '=':
			return 0;
		    case '+':
		    case '-':
		    case '*':
		    case '/':
			return -1;
		    default:
			printf("irregal char %c in compare.",beta);
			exit(1);
		}
		break;
	    case '+':
	    case '-':
		switch (beta) {
		    case '=':
			return 1;
		    case '+':
		    case '-':
			return 0;
		    case '*':
		    case '/':
			return -1;
		    default:
			printf("irregal char %c in compare.",beta);
			exit(1);
		}
		break;
	    case '*':
	    case '/':
		switch (beta) {
		    case '=':
		    case '+':
		    case '-':
			return 1;
		    case '*':
		    case '/':
			return 0;
		    default:
			printf("irregal char %c in compare.",beta);
			exit(1);
		}
		break;
	    default:
		printf("irregal char %c in compare.",alpha);
		exit(1);
		break;
	}
}

int conv_tree() /* tree conversion (if EOF return EOF) */
{
	struct word w,pw,jiku();
	struct node *np;
	char alpha,beta;
	int  flag;

	tree_root = NULL;

	for (;;) {
	    w = jiku();
#ifdef DEBUG
	    printf("jiku() = ");printword(w); printf("\n");
#endif
	    switch (w.type) {
		case VARNAME:
		case CONST:
		    np = makenode(w);
		    tree_s1push(np);
		    break;
		case SYMBOL:
		    alpha = w.symbol;
		    switch (alpha) {
			case '+':
			case '-':
			case '*':
			case '/':
			case '=':
			    flag = TRUE;
			    while (flag==TRUE) {
				if (tree_s2top == -1) {
				    tree_s2push((char)w.symbol);
				    flag = FALSE;
				} else if (tree_s2[tree_s2top]=='(') {
				    tree_s2push((char)w.symbol);
				    flag = FALSE;
				} else {
				    beta = tree_s2[tree_s2top];
				    switch (compare(alpha,beta)) {
					case -1: /* alpha < beta */
					case 0:  /* alpha = beta */
					    pw.symbol = tree_s2pop();
					    pw.type = SYMBOL;
					    np = makenode(pw);
					    np->right = tree_s1pop();
					    np->left = tree_s1pop();
					    tree_s1push(np);
					    flag = TRUE;
					    break;
					case 1: /* alpha > beta */
					    tree_s2push((char)w.symbol);
					    flag = FALSE;
					    break;
				    }
				}
			    }
			    break;
			case '(':
			    tree_s2push((char)w.symbol);
			    break;
			case '\n':
			case ')':
			    flag = TRUE;
			    while (flag==TRUE) {
				if (tree_s2[tree_s2top]=='(') {
				    tree_s2pop();
				    flag = FALSE;
				} else {
				    beta = tree_s2pop();
				    switch(beta) {
					case '+':
					case '-':
					case '*':
					case '/':
					    pw.symbol = beta;
					    pw.type = SYMBOL;
					    np = makenode(pw);
					    np->right = tree_s1pop();
					    np->left = tree_s1pop();
					    tree_s1push(np);
					    flag = TRUE;
					    break;
					case '=':
					    pw.symbol = beta;
					    pw.type = SYMBOL;
					    np = makenode(pw);
					    np->right = tree_s1pop();
					    np->left = tree_s1pop();
					    tree_root = np;
					    line++;
					    return 0;
					    break;
				    }
				}
			    }
			    break;
			case EOF:
			    return EOF;
			    break;
		    }
		    break;
		case ERROR:
		    error("irregal character");
		    return;
	    }
	}
}
