/* architecture simulator ver.1.0 [92/Aug/26-] */
/* debugger procedure */

#include <curses.h>
#include <string.h>
#include <ctype.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"
#include "display.h"
#include "debugger.h"

static	char	dumpfile[256]; /* dump file name */

debugger_main()
{
	char	cmdbuff[128];
	char	command[128];
	char	parameter[128];
	int	endstate;
	char	ch;
	int	address;
	int	i,tmp;
	int	dumpstart,dumpend;
	extern char wait_inkey();
	extern int progload();dumpmem();

	initial();		/* initialize simulator */
	clearbreakpoint();	/* clear breakpoint */
	endstate = HALT_END;	/* initialize end status */

	display_cyclephase(cycle);
        display_status();
	dispflush();
	mvprintw(COMMANDLINE+1,0,"load object filename : %s",objfile);
	progload(objfile); 
        display_status();
	dispflush();
	for(;;) {
	    mvinput(COMMANDLINE,0,"COMMAND: ",cmdbuff);
	    clearline(COMMANDLINE+1);
	    clearline(COMMANDLINE+2);
	    clearline(COMMANDLINE+3);

	    cmdlineanalyze(cmdbuff,command,parameter);
	    switch (checkcommand(command)) {
		case CMD_RUN:	/* initialize and run program */
		    initial();
		    if (parameter[0] != NULL) {
			PC = (int)strtol(parameter,(char **)NULL,0);
			mvprintw(COMMANDLINE+1,0,"initial PC = 0x%X(%d)",
						  PC & 0xffff, PC);
			dispflush();
		    }
		    endstate = simulate();
		    disp_endstate(endstate);
		    display_status();
		    dispflush();
		    break;
		case CMD_INIT:	/* initialize simulator */
		    initial();
		    mvprintw(COMMANDLINE+1,0,"initialize finished.");
		    dispflush();
		    break;
		case CMD_CONT:	/* continue stopped execution */
		    if ((endstate == BREAK_STOP)||(endstate == INTERRUPT)) {
		        endstate = simulate();
		        disp_endstate(endstate);
		        display_status();
		    } else { /* can't continue */
			mvprintw(COMMANDLINE+1,0,"can't continue.");
		    }
		    dispflush();
		    break;
#ifdef RISC_ARCH
		case CMD_DISPREG:	/* display registers */
		    dispreg(regdisptop);
		    dispflush();
		    while ((ch = wait_inkey()) != 'q') {
			switch (ch) {
			    case 'j':	/* display next 2 registers */
				regdisptop += 2;
				break;
			    case 'n':	/* display next 20 registers */
				regdisptop += 20;
				break;
			    case 'k':	/* display previous 2 registers */
				regdisptop -= 2;
				break;
			    case 'p':	/* display previous 20 registers */
				regdisptop -= 20;
				break;
			    default:
				break;
			}
			if (regdisptop < 0) { /* regdisptop range check */
			    regdisptop = 0;
			} else if (regdisptop > REGSIZE - 20) {
			    regdisptop = REGSIZE - 20;
			}
			dispreg(regdisptop);
			dispflush();
		    }
		    break;
#endif
		case CMD_DISPMEM:	/* display memories */
		    dispmem(memdisptop);
		    dispflush();
		    while ((ch = wait_inkey()) != 'q') {
			switch (ch) {
			    case 'j':	/* display next 2 memories */
				memdisptop += 2;
				break;
			    case 'n':	/* display next 20 memories */
				memdisptop += 20;
				break;
			    case 'k':	/* display previous 2 memories */
				memdisptop -= 2;
				break;
			    case 'p':	/* display previous 20 memories */
				memdisptop -= 20;
				break;
			    default:
				break;
			}
			if (memdisptop < 0) { /* memdisptop range check */
			    memdisptop = 0;
			} else if (memdisptop > MEMSIZE - 20) {
			    memdisptop = MEMSIZE - 20;
			}
			dispmem(memdisptop);
			dispflush();
		    }
		    break;
		case CMD_BREAKPOINT:	/* set breakpoint */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input breakpoint address : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"input breakpoint address : %s",parameter);
			clrtoeol();
		    }
		    address = (int) strtol(parameter,(char **)NULL,0);
		    setbreakpoint(address);
		    mvprintw(COMMANDLINE+2,0,"breakpoint is set at 0x%04X(%d).",
			     address & 0xffff,address);
		    dispflush();
		    break;
		case CMD_CLEARBP:	/* clear breakpoint*/
		    clearbreakpoint();
		    mvprintw(COMMANDLINE+1,0,"breakpoint is cleared.");
		    dispflush();
		    break;
		case CMD_WAIT:	/* set wait time */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input wait time (msec) : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"input wait time (msec) : %s",parameter);
			clrtoeol();
		    }
		    tmp = waittime;
		    waittime = (int) strtol(parameter,(char **)NULL,0) * 1000;
		    if (waittime <= 0) waittime = tmp;
		    mvprintw(COMMANDLINE+2,0,"wait time is set to %dmsec.",
			     (int)(waittime/1000));
		    dispflush();
		    break;
		case CMD_STEP:	/* set step execution mode */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"step execution on/off : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"step execution on/off : %s",parameter);
			clrtoeol();
		    }
		    if (strcmp(parameter,"on") == 0) {
			stepexec = TRUE;
			mvprintw(COMMANDLINE+2,0,"step mode ON.");
		    } else if (strcmp(parameter,"off") == 0) {
			stepexec = FALSE;
			mvprintw(COMMANDLINE+2,0,"step mode OFF.");
		    } else {  /* irregal input */
			mvprintw(COMMANDLINE+2,0,"step mode %s.",
				 (stepexec == TRUE)?"ON":"OFF");
		    }
		    dispflush();
		    break;
		case CMD_LOAD:	/* load object file */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input object filename : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"load object filename : %s",parameter);
			clrtoeol();
		    }
		    if (progload(parameter) == TRUE) { /* if load success */
			initial();		/* initialize simulator */
			clearbreakpoint();	/* clear breakpoint */
			endstate = HALT_END;/* initialize end status */
			display_cyclephase(cycle,0.0);
			display_status();
		    }
		    dispflush();
		    break;
		case CMD_DUMPMEM:	/* dump memory to file */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input memory dump filename : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"memory dump filename : %s",parameter);
			clrtoeol();
		    }
		    strcpy(dumpfile,parameter);
		    mvinput(COMMANDLINE+2,0,"input dump start address : ",parameter);
		    dumpstart = (int) strtol(parameter,(char **)NULL,0);
		    mvinput(COMMANDLINE+2,0,"input dump end address : ",parameter);
		    dumpend = (int) strtol(parameter,(char **)NULL,0);
		    if (dumpmem(dumpfile,dumpstart,dumpend)==TRUE) { /* dump success */
			mvprintw(COMMANDLINE+3,0,"memory[0x%X-0x%X] successfully dumped to the file '%s'.",
				 dumpstart,dumpend,dumpfile);
		    } else {	/* dump failed */
			mvprintw(COMMANDLINE+3,0,"memory dump failed.");
		    }
		    dispflush();
		    break;
		case CMD_STATUS:/* simulator status display */
		    clear();	/* clear screen */
		    mvprintw(COMMANDLINE+1,0,"STEP %s, wait time %d msec.",
			     (stepexec == TRUE)?"ON":"OFF" ,
			     (int)(waittime/1000)) ;
		    if (breakpoint.enable == TRUE) {
			mvprintw(COMMANDLINE+2,0,"breakpoint = 0x%04X(%d).",
				 breakpoint.address & 0xffff,breakpoint.address);
		    }
		    display_cyclephase(cycle,0.0);
		    display_status();
		    dispflush();
		    break;
		case CMD_QUIT:	/* quit debugger */
		    /*
		    mvinput(COMMANDLINE+1,0,"hit return to quit simulator.",parameter);
		    */
		    dispflush();
		    return;
		    break;
		case CMD_ERROR:
		default:
		    mvprintw(COMMANDLINE+1,0,"Command error.");
		    mvprintw(COMMANDLINE+2,0,"Commands : ");
		    i = 0;
		    while (cmdtable[i].type != CMD_ERROR) {
			printw("%s ",cmdtable[i].str);
			    i++;
		    }
		    dispflush();
		    break;
	    }
	}
}

int checkcommand(cmd)	/* check command and return command type */
char	*cmd;	/* command string */
{
	int	i,lastmatch,matchcount;

	/* perfect match test */
	i = 0;
	while (cmdtable[i].type != CMD_ERROR) {
		if (strcmp(cmd,cmdtable[i].str) == 0) {
			return (cmdtable[i].type);
		}
		i++;
	}

	/* substring match test */
	i = 0;
	matchcount = 0;
	while (cmdtable[i].type != CMD_ERROR) {
		if (strncmp(cmd,cmdtable[i].str,strlen(cmd)) == 0) {
			lastmatch = i;
			matchcount ++;
		}
		i++;
	}
	if (matchcount == 1) {
		return (cmdtable[lastmatch].type);
	}

	return (CMD_ERROR);	/* command string no match */
}

cmdlineanalyze(buff,cmd,param)	/* command line analyzer */
char	*buff;	/* input buffer */
char	*cmd;	/* command */
char	*param;	/* parameter */
{
	char	*cp;
	char	c;
	int	i;

	if (strchr(buff,(int)' ') == NULL) { /* if space is not found in buff */
	    strcpy(cmd,buff);
	    strcpy(param,"");
	} else {
	    cp = strchr(buff,(int)' ');
	    i = cp - buff;
	    strncpy(cmd,buff,i);
	    cmd[i] = (char) NULL;
	    strcpy(param,cp+1);
	}
	return;
}

mvinput(y,x,msg,str)	/* print message and get string */
int	y,x;	/* display message coordinates */
char	*msg;	/* message string */
char	*str;	/* input string */
{
	mvprintw(y,x,msg);
	clrtoeol();
	dispflush();
	getstrwithbs(str);
}

getstrwithbs(str)	/* get string which can use back-spacing */
char	*str;
{
	char	ch;
	char	buff[128];	/* input buffer */
	int	x,y;
	int	cp;

	cp = 0;
	buff[cp] = (char)0;

	getyx(stdscr,y,x);
	while((ch = wait_inkey())!='\n') {
	    if((ch >= ' ')&&(ch<='~')) { /* ch = space(0x20) - '~' (0x7e) */
		if (cp < 127) {		 /* append one char */
		    buff[cp] = ch;
		    addch(ch);
		    dispflush();
		    cp++;
		    buff[cp] = (char)0;
		}
	    } else if ((ch=='\b')||((int)ch==0x7f)) {	/* ch = backspace,DEL */
		if (cp > 0) {				/* delete one char */
		    ch = '\b';
		    addch(ch);
		    addch(' ');
		    addch(ch);
		    dispflush();
		    cp--;
		    buff[cp] = (char)0;
		}
	    } else if ((int)ch==0x1b) { /* ch = escape(0x1b) */
		move(y,x);		/* cancel all input characters */
		clrtoeol();
		dispflush();
		cp=0;
		buff[cp] = (char)0;
	    }
	}

	strcpy(str,buff);
}

char wait_inkey()	/* get one char with wait (no echo to screen) */
{
	char	ch;

	cbreak();	/* set cbreak mode */
	ch = getch();
	nocbreak();	/* set nocbreak mode */
	return ch;
}

void interrupt()	/* keyboard interrupt handler */
{
	interruptflag = TRUE; /* interrupt flag enable */
	return;
}

setbreakpoint(addr)	/* set breakpoint */
int	addr;	/* breakpoint address */
{
	breakpoint.enable = TRUE;
	breakpoint.address = addr;
	breakpoint.count = 0;
	breakpoint.limit = 0;
}

clearbreakpoint()	/* clear breakpoint */
{
	breakpoint.enable = FALSE;
	breakpoint.address = 0;
	breakpoint.count = 0;
	breakpoint.limit = 0;
}

int checkbreakpoint(pc) /* check breakpoint */
int	pc;	/* program counter value */
{
	if ((breakpoint.enable == TRUE) && (breakpoint.address == pc)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

disp_endstate(endstate)	/* display program stop status */
int	endstate;	/* end state */
{
	switch (endstate) {
	    case HALT_END:
		mvprintw(COMMANDLINE+1,0,"program is terminated by HALT instruction.");
		clrtoeol();
		break;
	    case ERROR_END:
		mvprintw(COMMANDLINE+1,0,"program is terminated by Run-time ERROR.");
		clrtoeol();
		break;
	    case BREAK_STOP:
		mvprintw(COMMANDLINE+1,0,"program is stopped at BREAKPOINT.");
		clrtoeol();
		break;
	    case INTERRUPT:
		mvprintw(COMMANDLINE+1,0,"program is stopped by INTERRUPT.");
		clrtoeol();
		break;
	    default:
		mvprintw(COMMANDLINE+1,0,"program is stopped with irregal endstate.");
		clrtoeol();
		break;
	}
}
