/* accumlator architecture simulator ver.1.0 [92/Aug/26-7/29] */
/* instruction execution */

#include <stdio.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"

void execinst(inst)	/* execute instruction */
decoded_inst	*inst;	/* decoded instruction */
{
	void	exec_loadstore();
	void	exec_intcalc2();
	void	exec_intcalc1();
	void	exec_misc();

	switch (inst->optype) {
	    case LOADSTORE:
			exec_loadstore(inst);
			break;
	    case INTCALC2:
			exec_intcalc2(inst);
			break;
	    case INTCALC1:
			exec_intcalc1(inst);
			break;
	    case MISC:
			exec_misc(inst);
			break;
	    default:
			error("undefined instruction in execinst");
			break;
	}
}

void exec_loadstore(inst) /* execute loadstore instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	switch (inst->subtype) {
	    case LS_LOAD:	/* LD, LDI */
			if (inst->src1flag == MEMOPR) { /* LD */
			    ACC = memory[inst->src1];
			} else {			/* LDI */
			    ACC = inst->src1;
			}
			break;
	    case LS_STORE:	/* ST */
			if (inst->dstnflag == MEMOPR) {	/* ST */
			    memory[inst->dstn] = ACC;
			} else {			/* error */
			    error("undefined instruction in exec_loadstore");
			}
			break;
	    default:
			error("undefined instruction in exec_loadstore");
			break;
	}
}

void exec_intcalc2(inst) /* execute integer op. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	word	operand;

	if (inst->src1flag == MEMOPR) { /* memory operand */
	    operand = memory[inst->src1];
	} else {			/* immediate operand */
	    operand = inst->src1;
	}

	switch (inst->subtype) {
	    case OP_ADD:	/* ADD, ADDI */
			ACC = ACC + operand;
			break;
	    case OP_SUB:	/* SUB, SUBI */
			ACC = ACC - operand;
			break;
	    case OP_MUL:	/* MUL, MULI */
			ACC = ACC * operand;
			break;
	    case OP_DIV:	/* DIV, DIVI */
			if (operand != 0) {
			    ACC = ACC / operand;
			} else { /* divide by 0 runtime error */
			    error("devide by 0 in exec_intcalc2");
			}
			break;
	    case OP_AND:	/* AND, ANDI */
			ACC = ACC & operand;
			break;
	    case OP_OR:		/* OR, ORI */
			ACC = ACC | operand;
			break;
	    case OP_XOR:	/* XOR, XORI */
			ACC = ACC ^ operand;
			break;
	    case OP_MOD:	/* MOD, MODI */
			if (operand != 0) {
			    ACC = ACC % operand;
			} else { /* divide by 0 runtime error */
			    error("devide by 0 in exec_intcalc2");
			}
			break;
	    default:
			error("undefined instruction in exec_intcalc2");
			break;
	}
}

void exec_intcalc1(inst) /* execute unary op. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	switch (inst->subtype) {
	    case UOP_NEG:	/* NEG */
			ACC = -ACC;
			break;
	    case UOP_NOT:	/* NOT */
			ACC = ~ACC;
			break;
	    default:
			error("undefined instruction in exec_intcalc1");
			break;
	}
}

void exec_misc(inst) /* execute misc. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	switch (inst->subtype) {
	    case MISC_NOP:	/* NOP */
			/* do nothing */
			break;
	    case MISC_HLT:	/* HLT */
			haltflag = TRUE;
			break;
	    default:
			error("undefined instruction in exec_misc");
			break;
	}
}
