/* cisc-like architecture simulator ver.1.0 [92/Aug/27-] */
/* instruction disassemble */

#include <stdio.h>
#include <string.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"
#include "mnemonic.h"

static int bitmask[]={0,     0x1,       0x3,       0x7,       0xf,
			    0x1f,      0x3f,      0x7f,      0xff,
		           0x1ff,     0x3ff,     0x7ff,     0xfff,
		          0x1fff,    0x3fff,    0x7fff,    0xffff,
		         0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
		        0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
		       0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
		      0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

#define GETBIT(var,pos,width) (((var) >> (pos)) & bitmask[(width)])

word disasm(pc,line)	/* disassemble instruction , return next pc */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;	/* current instruction */
	word	nextpc;		/* next pc */
	word	disasm_loadstore();
	word	disasm_intcalc2();
	word	disasm_intcalc1();
	word	disasm_misc();

	curinst = memory [pc];

	switch (GETBIT(curinst,13,3)) {
	    case LOADSTORE:
			nextpc = disasm_loadstore(pc,line);
			break;
	    case INTCALC2:
			nextpc = disasm_intcalc2(pc,line);
			break;
	    case INTCALC1:
			nextpc = disasm_intcalc1(pc,line);
			break;
	    case MISC:
			nextpc = disasm_misc(pc,line);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm");
			nextpc = pc;
			break;
	}
	return (nextpc); /* return next pc */
}

word disasm_loadstore(pc,line) /* disassemble loadstore instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst,curinst2;
	int	subtype,src1flag,src2flag,dstnflag;

	curinst = memory[pc++];	/* fetch instruction 1st word */
	curinst2 = memory[pc++];/* fetch instruction 2nd word */

	subtype = GETBIT(curinst,9,4);
	src1flag = GETBIT(curinst,6,2);
	dstnflag = GETBIT(curinst2,6,2);

	switch (subtype) {
	    case LS_LOAD:	/* LD [MI],R / MOV R,R */
		    if (dstnflag == REGOPR) {
			switch (src1flag) {
			    case MEMOPR:	/* LD M,R */
				sprintf(line,
					"%s	%04XH,R%d",
					mnemonic[LOADSTORE][subtype][0],
					memory[pc++] & 0xffff,
					GETBIT(curinst2,0,6)
					);
				break;
			    case IMMOPR:	/* LD I,R */
				sprintf(line,
					"%s	#%04XH,R%d",
					mnemonic[LOADSTORE][subtype][0],
					memory[pc++] & 0xffff,
					GETBIT(curinst2,0,6)
					);
				break;
			    case REGOPR:	/* MOV I,R */
				sprintf(line,
					"%s	R%d,R%d",
					mnemonic[LOADSTORE][subtype][1],
					memory[pc++] & 0xffff,
					GETBIT(curinst2,0,6)
					);
				break;
			    default:
				strcpy(line,"(undefined instruction)");
				error("undefined instruction in disasm_loadstore");
				pc-=2;	/* pc decrement to point error inst. */
				break;
			}
		    } else { /* destination isn't register */
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_loadstore");
			pc-=2;	/* pc decrement to point error inst. */
		    }
		    break;
	    case LS_STORE:	/* ST */
		    if (dstnflag == MEMOPR) {
			switch (src1flag) {
			    case MEMOPR:	/* MOV M,M */
				sprintf(line,
					"%s	%04XH,%04XH",
					mnemonic[LOADSTORE][subtype][1],
					memory[pc  ] & 0xffff,
					memory[pc+1] & 0xffff
					);
				pc += 2;
				break;
			    case IMMOPR:	/* ST I,M */
				sprintf(line,
					"%s	#%04XH,%04XH",
					mnemonic[LOADSTORE][subtype][0],
					memory[pc  ] & 0xffff,
					memory[pc+1] & 0xffff
					);
				pc += 2;
				break;
			    case REGOPR:	/* ST R,M */
				sprintf(line,
					"%s	R%d,%04XH",
					mnemonic[LOADSTORE][subtype][0],
					GETBIT(curinst,0,6),
					memory[pc++] & 0xffff
					);
				break;
			    default:
				strcpy(line,"(undefined instruction)");
				error("undefined instruction in disasm_loadstore");
				pc-=2;	/* pc decrement to point error inst. */
				break;
			}
		    } else { /* destination isn't memory */
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_loadstore");
			pc-=2;	/* pc decrement to point error inst. */
		    }
		    break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_loadstore");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_intcalc2(pc,line) /* disassemble integer op. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst,curinst2;
	int	subtype,src1flag,src2flag,dstnflag;
	static char src1[10],src2[10],dstn[10];
	word	orgpc;

	orgpc = pc;		/* save original pc */

	curinst = memory[pc++];	/* fetch instruction 1st word */
	curinst2 = memory[pc++];	/* fetch instruction 2nd word */

	subtype = GETBIT(curinst,9,4);
	src1flag = GETBIT(curinst,6,2);
	src2flag = GETBIT(curinst2,14,2);
	dstnflag = GETBIT(curinst2,6,2);

	switch (subtype) {
	    case OP_ADD:	/* ADD, ADDI */
	    case OP_SUB:	/* SUB, SUBI */
	    case OP_MUL:	/* MUL, MULI */
	    case OP_DIV:	/* DIV, DIVI */
	    case OP_AND:	/* AND, ANDI */
	    case OP_OR:		/* OR, ORI */
	    case OP_XOR:	/* XOR, XORI */
	    case OP_MOD:	/* MOD, MODI */
			switch (src1flag) {
			    case REGOPR:
				    sprintf(src1,
					    "R%d",
					    GETBIT(curinst,0,6)
					    );
				    break;
			    case MEMOPR:
				    sprintf(src1,
					    "%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    case IMMOPR:
				    sprintf(src1,
					    "#%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    default:
				    strcpy(line,"(undefined instruction)");
				    error("undefined instruction in disasm_intcalc2");
				    /* restore original pc to point error inst. */
				    pc=orgpc;
				    break;
			}
			switch (src2flag) {
			    case REGOPR:
				    sprintf(src2,
					    "R%d",
					    GETBIT(curinst2,8,6)
					    );
				    break;
			    case MEMOPR:
				    sprintf(src2,
					    "%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    case IMMOPR:
				    sprintf(src2,
					    "#%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    default:
				    strcpy(line,"(undefined instruction)");
				    error("undefined instruction in disasm_intcalc2");
				    /* restore original pc to point error inst. */
				    pc=orgpc;
				    break;
			}
			switch (dstnflag) {
			    case REGOPR:
				    sprintf(dstn,
					    "R%d",
					    GETBIT(curinst2,0,6)
					    );
				    break;
			    case MEMOPR:
				    sprintf(dstn,
					    "%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    default:
				    strcpy(line,"(undefined instruction)");
				    error("undefined instruction in disasm_intcalc2");
				    /* restore original pc to point error inst. */
				    pc=orgpc;
				    break;
			}
			if (errorflag == FALSE) {
			    sprintf(line,"%s	%s,%s,%s",
					mnemonic[INTCALC2][subtype][0],
					src1,src2,dstn);
			}
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_intcalc2");
			/* restore original pc to point error inst. */
			pc=orgpc;
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_intcalc1(pc,line) /* disassemble unary op. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst,curinst2;
	int	subtype,src1flag,dstnflag;
	static char src1[10],dstn[10];
	word	orgpc;

	orgpc = pc;		/* save original pc */

	curinst = memory[pc++];	/* fetch instruction 1st word */
	curinst2 = memory[pc++];/* fetch instruction 2nd word */

	subtype = GETBIT(curinst,9,4);
	src1flag = GETBIT(curinst,6,2);
	dstnflag = GETBIT(curinst2,6,2);

	if ((dstnflag == REGOPR)||(dstnflag == MEMOPR)) {
	    switch (subtype) {
		case UOP_NEG:	/* NEG */
		case UOP_NOT:	/* NOT */
			switch (src1flag) {
			    case REGOPR:
				    sprintf(src1,
					    "R%d",
					    GETBIT(curinst,0,6)
					    );
				    break;
			    case MEMOPR:
				    sprintf(src1,
					    "%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    case IMMOPR:
				    sprintf(src1,
					    "#%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    default:
				    strcpy(line,"(undefined instruction)");
				    error("undefined instruction in disasm_intcalc1");
				    /* original pc restore to point error inst. */
				    pc=orgpc;
				    break;
			}
			switch (dstnflag) {
			    case REGOPR:
				    sprintf(dstn,
					    "R%d",
					    GETBIT(curinst2,0,6)
					    );
				    break;
			    case MEMOPR:
				    sprintf(dstn,
					    "%04XH",
					    memory[pc++] & 0xffff
					    );
				    break;
			    default:
				    strcpy(line,"(undefined instruction)");
				    error("undefined instruction in disasm_intcalc1");
				    /* original pc restore to point error inst. */
				    pc=orgpc;
				    break;
			}

			sprintf(line,"%s	%s,%s",
				mnemonic[INTCALC1][subtype][0],
				src1, dstn);
		        break;
		default:
			    strcpy(line,"(undefined instruction)");
			    error("undefined instruction in disasm_intcalc1");
			    /* original pc restore to point error inst. */
			    pc=orgpc;
			    break;
	    }
	} else {
	    strcpy(line,"(undefined instruction)");
	    error("undefined instruction in disasm_intcalc1");
	    /* original pc restore to point error inst. */
	    pc=orgpc;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_misc(pc,line) /* disassemble misc. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;
	int	subtype;

	curinst = memory[pc++];

	subtype = GETBIT(curinst,9,4);

	switch (subtype) {
	    case MISC_NOP:	/* NOP */
			sprintf(line,"%s",
				    mnemonic[MISC][subtype][0]);
			break;
	    case MISC_HLT:	/* HLT */
			sprintf(line,"%s",
				    mnemonic[MISC][subtype][0]);
			pc--;	/* pc decrement to point HALT inst. */
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_misc");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}
