/* cisc-like architecture simulator ver.1.0 [92/Aug/27-] */
/* instruction execution */

#include <stdio.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"

void execinst(inst)	/* execute instruction */
decoded_inst	*inst;	/* decoded instruction */
{
	void	exec_loadstore();
	void	exec_intcalc2();
	void	exec_intcalc1();
	void	exec_misc();

	switch (inst->optype) {
	    case LOADSTORE:
			exec_loadstore(inst);
			break;
	    case INTCALC2:
			exec_intcalc2(inst);
			break;
	    case INTCALC1:
			exec_intcalc1(inst);
			break;
	    case MISC:
			exec_misc(inst);
			break;
	    default:
			error("undefined instruction in execinst");
			break;
	}
}

void exec_loadstore(inst) /* execute loadstore instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	word	operand;

	switch (inst->subtype) {
	    case LS_LOAD:	/* LD [MI],R / MOV R,R */
		    switch (inst->src1flag) {
			case REGOPR:	/* MOV R,R */
				operand = regfile[inst->src1];
				break;
			case MEMOPR:	/* LD M,R */
				operand = memory[inst->src1];
				break;
			case IMMOPR:	/* LD I,R */
				operand = inst->src1;
				break;
			default:
				error("undefined instruction in exec_loadstore");
				break;
		    }
		    if (inst->dstnflag == REGOPR) {
			regfile[inst->dstn] = operand;
		    } else {			/* error */
			error("undefined instruction in exec_loadstore");
		    }
		    break;
	    case LS_STORE:	/* ST [RI],M / MOV M,M */
		    switch (inst->src1flag) {
			case REGOPR:	/* ST R,M */
				operand = regfile[inst->src1];
				break;
			case MEMOPR:	/* MOV M,M */
				operand = memory[inst->src1];
				break;
			case IMMOPR:	/* ST I,M */
				operand = inst->src1;
				break;
			default:
				error("undefined instruction in exec_loadstore");
				break;
		    }
		    if (inst->dstnflag == MEMOPR) {
			memory[inst->dstn] = operand;
		    } else {			/* error */
			error("undefined instruction in exec_loadstore");
		    }
		    break;
	    default:
		    error("undefined instruction in exec_loadstore");
		    break;
	}
}

void exec_intcalc2(inst) /* execute integer op. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	word	operand1,operand2,result;

	switch (inst->src1flag) {
	    case REGOPR:	/* register operand */
		    operand1 = regfile[inst->src1];
		    break;
	    case MEMOPR:	/* memory operand */
		    operand1 = memory[inst->src1];
		    break;
	    case IMMOPR:	/* immediate operand */
		    operand1 = inst->src1;
		    break;
	    default:
		    error("undefined instruction in exec_intcalc2");
		    break;
	}
	switch (inst->src2flag) {
	    case REGOPR:	/* register operand */
		    operand2 = regfile[inst->src2];
		    break;
	    case MEMOPR:	/* memory operand */
		    operand2 = memory[inst->src2];
		    break;
	    case IMMOPR:	/* immediate operand */
		    operand2 = inst->src2;
		    break;
	    default:
		    error("undefined instruction in exec_intcalc2");
		    break;
	}

	switch (inst->subtype) {
	    case OP_ADD:	/* ADD, ADDI */
			result = operand1 + operand2;
			break;
	    case OP_SUB:	/* SUB, SUBI */
			result = operand1 - operand2;
			break;
	    case OP_MUL:	/* MUL, MULI */
			result = operand1 * operand2;
			break;
	    case OP_DIV:	/* DIV, DIVI */
			if (operand2 != 0) {
			    result = operand1 / operand2;
			} else { /* divide by 0 runtime error */
			    result = 0;
			    error("devide by 0 in exec_intcalc2");
			}
			break;
	    case OP_AND:	/* AND, ANDI */
			result = operand1 & operand2;
			break;
	    case OP_OR:		/* OR, ORI */
			result = operand1 | operand2;
			break;
	    case OP_XOR:	/* XOR, XORI */
			result = operand1 ^ operand2;
			break;
	    case OP_MOD:	/* MOD, MODI */
			if (operand2 != 0) {
			    result = operand1 % operand2;
			} else { /* divide by 0 runtime error */
			    result = 0;
			    error("devide by 0 in exec_intcalc2");
			}
			break;
	    default:
			error("undefined instruction in exec_intcalc2");
			result = 0;
			break;
	}

	switch (inst->dstnflag) {
	    case REGOPR:	/* register operand */
		    regfile[inst->dstn] = result;
		    break;
	    case MEMOPR:	/* memory operand */
		    memory[inst->dstn] = result;
		    break;
	    case IMMOPR:	/* immediate operand */
	    default:
		    error("undefined instruction in exec_intcalc2");
		    break;
	}
}

void exec_intcalc1(inst) /* execute unary op. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	word	operand,result;

	switch (inst->src1flag) {
	    case REGOPR:	/* register operand */
		    operand = regfile[inst->src1];
		    break;
	    case MEMOPR:	/* memory operand */
		    operand = memory[inst->src1];
		    break;
	    case IMMOPR:	/* immediate operand */
		    operand = inst->src1;
		    break;
	    default:
		    error("undefined instruction in exec_intcalc1");
		    break;
	}

	switch (inst->subtype) {
	    case UOP_NEG:	/* NEG */
			result = -operand;
			break;
	    case UOP_NOT:	/* NOT */
			result = ~operand;
			break;
	    default:
			error("undefined instruction in exec_intcalc1");
			result = 0;
			break;
	}

	switch (inst->dstnflag) {
	    case REGOPR:	/* register operand */
		    regfile[inst->dstn] = result;
		    break;
	    case MEMOPR:	/* memory operand */
		    memory[inst->dstn] = result;
	    case IMMOPR:	/* immediate operand */
	    default:
		    error("undefined instruction in exec_intcalc2");
		    break;
	}
}

void exec_misc(inst) /* execute misc. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	switch (inst->subtype) {
	    case MISC_NOP:	/* NOP */
			/* do nothing */
			break;
	    case MISC_HLT:	/* HLT */
			haltflag = TRUE;
			break;
	    default:
			error("undefined instruction in exec_misc");
			break;
	}
}
