/* architecture simulator ver.1.0 [92/Aug./26-] */
/* header file for debugger procedures */

#include "global.h"

#define	HALT_END 0	/* terminated by HALT instruction */
#define	ERROR_END 1	/* terminated by Run-time ERROR */
#define BREAK_STOP 2	/* stopped by BREAKPOINT */
#define INTERRUPT 3	/* stopped by INTERRUPT */

#define	CMD_RUN	0	/* debugger RUN command */
#define	CMD_INIT 1	/* debugger INIT command */
#define	CMD_TRACE 2	/* debugger TRACE command */
#define	CMD_CONT 3	/* debugger CONT command */
#define	CMD_DISPREG 4	/* debugger DISPSR command */
#define	CMD_DISPMEM 5	/* debugger DISPMEM command */
#define	CMD_BREAKPOINT 6 /* debugger BREAKPOINT command */
#define	CMD_CLEARBP 7	/* debugger CLEARBP command */
#define	CMD_QUIT 8	/* debugger QUIT command */
#define	CMD_WAIT 9	/* debugger WAIT command */
#define	CMD_STEP 10	/* debugger STEP command */
#define	CMD_STATUS 11	/* debugger STATUS command */
#define	CMD_LOAD 12	/* debugger LOAD command */
#define	CMD_DUMPMEM 13	/* debugger DUMPMEM command */
#define	CMD_ERROR -1	/* illegal debugger command */

typedef struct {	/* struct for debugger commands */
	int	type;	/* command type number */
	char	*str;	/* command string */
} cmd_struct;

#ifdef MAINPROGRAM
cmd_struct cmdtable[]={
	{CMD_RUN,"run"},
	{CMD_INIT,"init"},
	/* {CMD_TRACE,"trace"}, */	/* trace command not used */
	{CMD_CONT,"cont"},
#ifdef RISC_ARCH
	{CMD_DISPREG,"dreg"},
#endif
	{CMD_DISPMEM,"dmem"},
	{CMD_BREAKPOINT,"bp"},
	{CMD_CLEARBP,"clearbp"},
	{CMD_WAIT,"wait"},
	{CMD_STEP,"step"},
	{CMD_STATUS,"status"},
	{CMD_LOAD,"load"},
	{CMD_DUMPMEM,"dumpmem"},
	{CMD_QUIT,"quit"},
	{CMD_ERROR,""}
};
#else
extern	cmd_struct cmdtable[];
#endif

typedef struct {	/* struct for breakpoint */
	int	address;	/* breakpoint address */
	int	enable;		/* enable flag */
	int	count;		/* counter (for future extention) */
	int	limit;		/* limit count (for future extention) */
} bp_struct;

global bp_struct breakpoint;	/* breakpoint */

global int cycle init(0);	/* executed cycle */
global int interruptflag init(FALSE);	/* keybord interrupt flag */
global int stepexec init(TRUE);/* step execution flag */

global char objfile[256];	/* object file name */

extern int checkbreakpoint();	/* breakpoint checking routine */
extern void interrupt();	/* keybord interrupt handler */
