/* risc-like architecture simulator ver.1.0 [92/Aug/27-] */
/* instruction decode */

#include <stdio.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"

static int bitmask[]={0,     0x1,       0x3,       0x7,       0xf,
			    0x1f,      0x3f,      0x7f,      0xff,
		           0x1ff,     0x3ff,     0x7ff,     0xfff,
		          0x1fff,    0x3fff,    0x7fff,    0xffff,
		         0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
		        0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
		       0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
		      0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

#define GETBIT(var,pos,width) (((var) >> (pos)) & bitmask[(width)])

#define MERGE(hi,lo)	((((hi) & 0xffff) << 16) | ((lo) & 0xffff))

word decode(pc,inst)	/* decode instruction , return next PC */
word		pc;	/* current pc */
decoded_inst	*inst;	/* decoded instruction */
{
	word	curinst1,curinst2; /* current instruction */
	int	curinst;	/* current RISC instruction */
	word	nextpc;		/* next pc */
	void	decode_loadstore();
	void	decode_intcalc2();
	void	decode_intcalc1();
	void	decode_misc();

	curinst1 = memory [pc++];	/* fetch instruction (high) */
	curinst2 = memory [pc++];	/* fetch instruction (lo)   */

	curinst = MERGE(curinst1,curinst2); /* merge RISC instruction */

	inst->optype = GETBIT(curinst,30,2);
	switch (inst->optype) {
	    case LOADSTORE:
			decode_loadstore(curinst,inst);
			break;
	    case INTCALC2:
			decode_intcalc2(curinst,inst);
			break;
	    case INTCALC1:
			decode_intcalc1(curinst,inst);
			break;
	    case MISC:
			decode_misc(curinst,inst);
			break;
	    default:
			error("undefined instruction in decode");
			break;
	}
	if (errorflag == TRUE) {
	    pc-=2;		/* pc decrement to point error inst. */
	} else if (haltflag == TRUE) {
	    pc-=2;		/* pc decrement to point HALT inst. */
	}
	return (pc); /* return next pc */
}

void decode_loadstore(curinst,inst) /* decode loadstore instructions */
int		curinst;/* current instruction(32bit) */
decoded_inst	*inst;	/* decoded instruction */
{
	inst->subtype = GETBIT(curinst,27,3);

	switch (inst->subtype) {
	    case LS_LOAD:	/* LD, LDI */
			inst->src1flag = GETBIT(curinst,26,1);
			inst->dstn = GETBIT(curinst,0,5);
			inst->src1 = GETBIT(curinst,5,16);
			break;
	    case LS_STORE:	/* ST */
			inst->src1flag = GETBIT(curinst,26,1);
			if (inst->src1flag!=REGOPR) {
			    error("undefined instruction in decode_loadstore");
			} else {
			    inst->src1 = GETBIT(curinst,21,5);
			    inst->dstn = GETBIT(curinst,5,16);
			}
			break;
	    default:
			error("undefined instruction in decode_loadstore");
			break;
	}
}

void decode_intcalc2(curinst,inst) /* decode integer op. instructions */
int		curinst;/* current instruction(32bit) */
decoded_inst	*inst;	/* decoded instruction */
{
	inst->subtype = GETBIT(curinst,27,3);

	switch (inst->subtype) {
	    case OP_ADD:	/* ADD, ADDI */
	    case OP_SUB:	/* SUB, SUBI */
	    case OP_MUL:	/* MUL, MULI */
	    case OP_DIV:	/* DIV, DIVI */
	    case OP_AND:	/* AND, ANDI */
	    case OP_OR:		/* OR, ORI */
	    case OP_XOR:	/* XOR, XORI */
	    case OP_MOD:	/* MOD, MODI */
			inst->src2flag = GETBIT(curinst,26,1);
			inst->src1 = GETBIT(curinst,21,5);
			inst->src2 = GETBIT(curinst,5,16);
			inst->dstn = GETBIT(curinst,0,5);
			break; default:
			error("undefined instruction in decode_intcalc2");
			break;
	}
}

void decode_intcalc1(curinst,inst) /* decode unary op. instructions */
int		curinst;/* current instruction(32bit) */
decoded_inst	*inst;	/* decoded instruction */
{
	inst->subtype = GETBIT(curinst,27,3);

	switch (inst->subtype) {
	    case UOP_NEG:	/* NEG */
	    case UOP_NOT:	/* NOT */
			inst->src1flag = GETBIT(curinst,26,1);
			if (inst->src1flag == REGOPR) {
			    /* register operand */
			    inst->src1 = GETBIT(curinst,21,5);
			} else {
			    /* immediate operand */
			    inst->src1 = GETBIT(curinst,5,16);
			}
			inst->dstn = GETBIT(curinst,0,5);
			break;
	    default:
			error("undefined instruction in decode_intcalc1");
			break;
	}
}

void decode_misc(curinst,inst) /* decode misc. instructions */
int		curinst;/* current instruction(32bit) */
decoded_inst	*inst;	/* decoded instruction */
{
	inst->subtype = GETBIT(curinst,27,3);

	switch (inst->subtype) {
	    case MISC_NOP:	/* NOP */
			/* do nothing */
			break;
	    case MISC_HLT:	/* HLT */
			haltflag = TRUE;
			break;
	    default:
			error("undefined instruction in decode_misc");
			break;
	}
}
