/* accumlator architecture simulator ver.1.0 [92/Aug/26-] */
/* instruction disassemble */

#include <stdio.h>
#include <string.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"
#include "mnemonic.h"

static int bitmask[]={0,     0x1,       0x3,       0x7,       0xf,
			    0x1f,      0x3f,      0x7f,      0xff,
		           0x1ff,     0x3ff,     0x7ff,     0xfff,
		          0x1fff,    0x3fff,    0x7fff,    0xffff,
		         0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
		        0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
		       0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
		      0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

#define GETBIT(var,pos,width) (((var) >> (pos)) & bitmask[(width)])

#define MERGE(hi,lo)    ((((hi) & 0xffff) << 16) | ((lo) & 0xffff))

word disasm(pc,line)	/* disassemble instruction , return next pc */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;	/* current instruction */
	word	nextpc;		/* next pc */
	word	disasm_loadstore();
	word	disasm_intcalc2();
	word	disasm_intcalc1();
	word	disasm_misc();

	curinst = memory [pc];

	switch (GETBIT(curinst,14,2)) {
	    case LOADSTORE:
			nextpc = disasm_loadstore(pc,line);
			break;
	    case INTCALC2:
			nextpc = disasm_intcalc2(pc,line);
			break;
	    case INTCALC1:
			nextpc = disasm_intcalc1(pc,line);
			break;
	    case MISC:
			nextpc = disasm_misc(pc,line);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm");
			nextpc = pc;
			break;
	}
	return (nextpc); /* return next pc */
}

word disasm_loadstore(pc,line) /* disassemble loadstore instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	int	curinst;
	word	curinst1,curinst2;
	int	subtype,immflag;

	curinst1 = memory[pc++];	/* fetch instruction (high) */
	curinst2 = memory[pc++];	/* fetch instruction (lo)   */
	curinst = MERGE(curinst1,curinst2); /* merge RISC instruction */

	subtype = GETBIT(curinst,27,3);
	immflag = GETBIT(curinst,26,1);

	switch (subtype) {
	    case LS_LOAD:	/* LD, LDI */
			sprintf(line,"%s	%04XH,R%d",
				    mnemonic[LOADSTORE][subtype][immflag],
				    GETBIT(curinst,5,16) & 0xffff ,
				    GETBIT(curinst,0,5));
			break;
	    case LS_STORE:	/* ST */
			sprintf(line,"%s	%04XH,R%d",
				    mnemonic[LOADSTORE][subtype][immflag],
				    GETBIT(curinst,5,16) & 0xffff ,
				    GETBIT(curinst,21,5));
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_loadstore");
			pc-=2;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_intcalc2(pc,line) /* disassemble integer op. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	int	curinst;
	word	curinst1,curinst2;
	int	subtype,immflag;

	curinst1 = memory[pc++];	/* fetch instruction (high) */
	curinst2 = memory[pc++];	/* fetch instruction (lo)   */
	curinst = MERGE(curinst1,curinst2); /* merge RISC instruction */

	subtype = GETBIT(curinst,27,3);
	immflag = GETBIT(curinst,26,1);

	switch (subtype) {
	    case OP_ADD:	/* ADD, ADDI */
	    case OP_SUB:	/* SUB, SUBI */
	    case OP_MUL:	/* MUL, MULI */
	    case OP_DIV:	/* DIV, DIVI */
	    case OP_AND:	/* AND, ANDI */
	    case OP_OR:		/* OR, ORI */
	    case OP_XOR:	/* XOR, XORI */
	    case OP_MOD:	/* MOD, MODI */
			if (immflag == IMMOPR) {/* immediate operand */
			    sprintf(line,
				    "%s	R%d,%04XH,R%d",
				    mnemonic[INTCALC2][subtype][immflag],
				    GETBIT(curinst,21,5),
				    GETBIT(curinst,5,16) & 0xffff,
				    GETBIT(curinst,0,5)
				    );
			} else {		/* register operand */
			    sprintf(line,
				    "%s	R%d,R%d,R%d",
				    mnemonic[INTCALC2][subtype][immflag],
				    GETBIT(curinst,21,5),
				    GETBIT(curinst,5,5),
				    GETBIT(curinst,0,5)
				    );
			}
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_intcalc2");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_intcalc1(pc,line) /* disassemble unary op. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	int	curinst;
	word	curinst1,curinst2;
	int	subtype;
	int	immflag;

	curinst1 = memory[pc++];	/* fetch instruction (high) */
	curinst2 = memory[pc++];	/* fetch instruction (lo)   */
	curinst = MERGE(curinst1,curinst2); /* merge RISC instruction */

	subtype = GETBIT(curinst,27,3);
	immflag = GETBIT(curinst,26,1);

	switch (subtype) {
	    case UOP_NEG:	/* NEG */
	    case UOP_NOT:	/* NOT */
			if (immflag == IMMOPR) {/* immediate operand */
			    sprintf(line,
				    "%s	%04XH,R%d",
				     mnemonic[INTCALC1][subtype][0],
				     GETBIT(curinst,5,16),
				     GETBIT(curinst,0,5)
				   );
			} else {		/* register operand */
			    sprintf(line,
				    "%s	R%d,R%d",
				     mnemonic[INTCALC1][subtype][0],
				     GETBIT(curinst,21,5),
				     GETBIT(curinst,0,5)
				   );
			}
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_intcalc1");
			pc-=2;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_misc(pc,line) /* disassemble misc. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	int	curinst;
	word	curinst1,curinst2;
	int	subtype;

	curinst1 = memory[pc++];	/* fetch instruction (high) */
	curinst2 = memory[pc++];	/* fetch instruction (lo)   */
	curinst = MERGE(curinst1,curinst2); /* merge RISC instruction */

	subtype = GETBIT(curinst,27,3);

	switch (subtype) {
	    case MISC_NOP:	/* NOP */
	    case MISC_HLT:	/* HLT */
			sprintf(line,"%s",
				    mnemonic[MISC][subtype][0]);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_misc");
			pc-=2;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}
