/* risc-like architecture simulator ver.1.0 [92/Aug/27-] */
/* instruction execution */

#include <stdio.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"

void execinst(inst)	/* execute instruction */
decoded_inst	*inst;	/* decoded instruction */
{
	void	exec_loadstore();
	void	exec_intcalc2();
	void	exec_intcalc1();
	void	exec_misc();

	switch (inst->optype) {
	    case LOADSTORE:
			exec_loadstore(inst);
			break;
	    case INTCALC2:
			exec_intcalc2(inst);
			break;
	    case INTCALC1:
			exec_intcalc1(inst);
			break;
	    case MISC:
			exec_misc(inst);
			break;
	    default:
			error("undefined instruction in execinst");
			break;
	}
}

void exec_loadstore(inst) /* execute loadstore instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	switch (inst->subtype) {
	    case LS_LOAD:	/* LD, LDI */
			if (inst->src1flag != IMMOPR) { /* LD */
			    regfile[inst->dstn] = memory[inst->src1];
			} else {			/* LDI */
			    regfile[inst->dstn] = inst->src1;
			}
			break;
	    case LS_STORE:	/* ST */
			if (inst->dstnflag != IMMOPR) {	/* ST */
			    memory[inst->dstn] = regfile[inst->src1];
			} else {			/* error */
			    error("undefined instruction in exec_loadstore");
			}
			break;
	    default:
			error("undefined instruction in exec_loadstore");
			break;
	}
}

void exec_intcalc2(inst) /* execute integer op. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	word	operand1,operand2,result;

	operand1 = regfile[inst->src1];
	if (inst->src2flag == REGOPR) { /* register operand */
	    operand2 = regfile[inst->src2];
	} else {			/* immediate operand */
	    operand2 = inst->src2;
	}

	switch (inst->subtype) {
	    case OP_ADD:	/* ADD, ADDI */
			result = operand1 + operand2;
			break;
	    case OP_SUB:	/* SUB, SUBI */
			result = operand1 - operand2;
			break;
	    case OP_MUL:	/* MUL, MULI */
			result = operand1 * operand2;
			break;
	    case OP_DIV:	/* DIV, DIVI */
			if (operand2 != 0) {
			    result = operand1 / operand2;
			} else { /* divide by 0 runtime error */
			    result = 0;
			    error("devide by 0 in exec_intcalc2");
			}
			break;
	    case OP_AND:	/* AND, ANDI */
			result = operand1 & operand2;
			break;
	    case OP_OR:		/* OR, ORI */
			result = operand1 | operand2;
			break;
	    case OP_XOR:	/* XOR, XORI */
			result = operand1 ^ operand2;
			break;
	    case OP_MOD:	/* MOD, MODI */
			if (operand2 != 0) {
			    result = operand1 % operand2;
			} else { /* divide by 0 runtime error */
			    result = 0;
			    error("devide by 0 in exec_intcalc2");
			}
			break;
	    default:
			result = 0;
			error("undefined instruction in exec_intcalc2");
			break;
	}
	regfile[inst->dstn] = result;
}

void exec_intcalc1(inst) /* execute unary op. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	word	operand,result;

	if (inst->src1flag == REGOPR) { /* register operand */
	    operand = regfile[inst->src1];
	} else {			/* immediate operand */
	    operand = inst->src1;
	}

	switch (inst->subtype) {
	    case UOP_NEG:	/* NEG */
			result = -operand;
			break;
	    case UOP_NOT:	/* NOT */
			result = ~operand;
			break;
	    default:
			result = 0;
			error("undefined instruction in exec_intcalc1");
			break;
	}
	regfile[inst->dstn] = result;
}

void exec_misc(inst) /* execute misc. instructions */
decoded_inst	*inst;	/* decoded instruction */
{
	switch (inst->subtype) {
	    case MISC_NOP:	/* NOP */
			/* do nothing */
			break;
	    case MISC_HLT:	/* HLT */
			haltflag = TRUE;
			break;
	    default:
			error("undefined instruction in exec_misc");
			break;
	}
}
