/* stack architecture simulator ver.1.0 [92/Aug/27-] */
/* instruction decode */

#include <stdio.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"

static int bitmask[]={0,     0x1,       0x3,       0x7,       0xf,
			    0x1f,      0x3f,      0x7f,      0xff,
		           0x1ff,     0x3ff,     0x7ff,     0xfff,
		          0x1fff,    0x3fff,    0x7fff,    0xffff,
		         0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
		        0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
		       0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
		      0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

#define GETBIT(var,pos,width) (((var) >> (pos)) & bitmask[(width)])

word decode(pc,inst)	/* decode instruction , return next PC */
word		pc;	/* current pc */
decoded_inst	*inst;	/* decoded instruction */
{
	word	curinst;	/* current instruction */
	word	nextpc;		/* next pc */
	word	decode_loadstore();
	word	decode_intcalc2();
	word	decode_intcalc1();
	word	decode_misc();

	curinst = memory [pc];

	inst->optype = GETBIT(curinst,13,3);
	switch (inst->optype) {
	    case LOADSTORE:
			nextpc = decode_loadstore(pc,inst);
			break;
	    case INTCALC2:
			nextpc = decode_intcalc2(pc,inst);
			break;
	    case INTCALC1:
			nextpc = decode_intcalc1(pc,inst);
			break;
	    case MISC:
			nextpc = decode_misc(pc,inst);
			break;
	    default:
			error("undefined instruction in decode");
			nextpc = pc;
			break;
	}
	return (nextpc); /* return next pc */
}

word decode_loadstore(pc,inst) /* decode loadstore instructions */
word		pc;	/* current pc */
decoded_inst	*inst;	/* decoded instruction */
{
	word	curinst;

	curinst = memory[pc++];

	inst->subtype = GETBIT(curinst,9,4);

	switch (inst->subtype) {
	    case LS_PUSH:	/* PUSH, PUSHI */
			inst->src1flag = GETBIT(curinst,8,1);
			inst->src1 = memory[pc++];
			break;
	    case LS_POP:	/* POP */
			inst->dstnflag = GETBIT(curinst,8,1);
			inst->dstn = memory[pc++];
			break;
	    default:
			error("undefined instruction in decode_loadstore");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word decode_intcalc2(pc,inst) /* decode integer op. instructions */
word		pc;	/* current pc */
decoded_inst	*inst;	/* decoded instruction */
{
	word	curinst;

	curinst = memory[pc++];

	inst->subtype = GETBIT(curinst,9,4);

	switch (inst->subtype) {
	    case OP_ADD:	/* ADD, ADDI */
	    case OP_SUB:	/* SUB, SUBI */
	    case OP_MUL:	/* MUL, MULI */
	    case OP_DIV:	/* DIV, DIVI */
	    case OP_AND:	/* AND, ANDI */
	    case OP_OR:		/* OR, ORI */
	    case OP_XOR:	/* XOR, XORI */
	    case OP_MOD:	/* MOD, MODI */
			/* do nothing */
			break;
	    default:
			error("undefined instruction in decode_intcalc2");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word decode_intcalc1(pc,inst) /* decode unary op. instructions */
word		pc;	/* current pc */
decoded_inst	*inst;	/* decoded instruction */
{
	word	curinst;

	curinst = memory[pc++];

	inst->subtype = GETBIT(curinst,9,4);

	switch (inst->subtype) {
	    case UOP_NEG:	/* NEG */
	    case UOP_NOT:	/* NOT */
			/* do nothing */
			break;
	    default:
			error("undefined instruction in decode_intcalc1");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word decode_misc(pc,inst) /* decode misc. instructions */
word		pc;	/* current pc */
decoded_inst	*inst;	/* decoded instruction */
{
	word	curinst;

	curinst = memory[pc++];

	inst->subtype = GETBIT(curinst,9,4);

	switch (inst->subtype) {
	    case MISC_NOP:	/* NOP */
			/* do nothing */
			break;
	    case MISC_HLT:	/* HLT */
			pc--;	/* pc decrement to point HALT inst. */
			break;
	    default:
			error("undefined instruction in decode_misc");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}
