/* stack architecture simulator ver.1.0 [92/Aug/27-] */
/* instruction disassemble */

#include <stdio.h>
#include <string.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"
#include "mnemonic.h"

static int bitmask[]={0,     0x1,       0x3,       0x7,       0xf,
			    0x1f,      0x3f,      0x7f,      0xff,
		           0x1ff,     0x3ff,     0x7ff,     0xfff,
		          0x1fff,    0x3fff,    0x7fff,    0xffff,
		         0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
		        0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
		       0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
		      0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

#define GETBIT(var,pos,width) (((var) >> (pos)) & bitmask[(width)])

word disasm(pc,line)	/* disassemble instruction , return next pc */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;	/* current instruction */
	word	nextpc;		/* next pc */
	word	disasm_loadstore();
	word	disasm_intcalc2();
	word	disasm_intcalc1();
	word	disasm_misc();

	curinst = memory [pc];

	switch (GETBIT(curinst,13,3)) {
	    case LOADSTORE:
			nextpc = disasm_loadstore(pc,line);
			break;
	    case INTCALC2:
			nextpc = disasm_intcalc2(pc,line);
			break;
	    case INTCALC1:
			nextpc = disasm_intcalc1(pc,line);
			break;
	    case MISC:
			nextpc = disasm_misc(pc,line);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm");
			nextpc = pc;
			break;
	}
	return (nextpc); /* return next pc */
}

word disasm_loadstore(pc,line) /* disassemble loadstore instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;
	int	subtype,immflag;

	curinst = memory[pc++];

	subtype = GETBIT(curinst,9,4);
	immflag = GETBIT(curinst,8,1);

	switch (subtype) {
	    case LS_PUSH:	/* PUSH, PUSHI */
	    case LS_POP:	/* POP */
			sprintf(line,"%s	%04XH",
				    mnemonic[LOADSTORE][subtype][immflag],
				    memory[pc++] & 0xffff);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_loadstore");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_intcalc2(pc,line) /* disassemble integer op. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;
	int	subtype,immflag;

	curinst = memory[pc++];

	subtype = GETBIT(curinst,9,4);
	immflag = GETBIT(curinst,8,1);

	switch (subtype) {
	    case OP_ADD:	/* ADD */
	    case OP_SUB:	/* SUB */
	    case OP_MUL:	/* MUL */
	    case OP_DIV:	/* DIV */
	    case OP_AND:	/* AND */
	    case OP_OR:		/* OR  */
	    case OP_XOR:	/* XOR */
	    case OP_MOD:	/* MOD */
			sprintf(line,"%s",
				    mnemonic[INTCALC2][subtype][immflag]);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_intcalc2");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_intcalc1(pc,line) /* disassemble unary op. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;
	int	subtype;

	curinst = memory[pc++];

	subtype = GETBIT(curinst,9,4);

	switch (subtype) {
	    case UOP_NEG:	/* NEG */
	    case UOP_NOT:	/* NOT */
			sprintf(line,"%s",
				    mnemonic[INTCALC1][subtype][0]);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_intcalc1");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}

word disasm_misc(pc,line) /* disassemble misc. instructions */
word	pc;	/* current pc */
char 	*line;	/* disassembled line */
{
	word	curinst;
	int	subtype;

	curinst = memory[pc++];

	subtype = GETBIT(curinst,9,4);

	switch (subtype) {
	    case MISC_NOP:	/* NOP */
	    case MISC_HLT:	/* HLT */
			sprintf(line,"%s",
				    mnemonic[MISC][subtype][0]);
			break;
	    default:
			strcpy(line,"(undefined instruction)");
			error("undefined instruction in disasm_misc");
			pc--;	/* pc decrement to point error inst. */
			break;
	}
	return  (pc) ;	/* return next pc */
}
