/* architecture simulator ver.1.0 [92/Aug./26-] */
/* display procedures */

#include <stdio.h>
#include <curses.h>

#include "defines.h"
#include "cmnvar.h"
#include "opcode.h"

#include "display.h"
#include "debugger.h"

static char *TFstr[]={"FALSE","TRUE"};

display(cycle)	/* display status and wait */
int cycle;
{
	char	ch;
	extern	char	wait_inkey();
	
	display_cyclephase(cycle);

	display_status();

	if (stepexec == FALSE) { /* continuous execution mode */
	    dispflush();	     /* flush display */
	    waitusec(waittime); /* waiting... */
	} else { /* step execution mode */
	    dispflush();	     /* flush display */
	    while ((ch = wait_inkey()) != '\n') {   /* wait for key input */
		if ((ch == 'q')||(ch == 'Q')) {
		    interruptflag = TRUE;	/* type 'q' then interrupt */
		}
	    }
	}
}

display_cyclephase(cycle) /* display cycle */
int	cycle;
{
	mvprintw (CYCLELINE,0,"cycle %8d : PC = %04X(%6d)",
		  cycle,PC & 0xFFFF,PC);
	clrtoeol();
}

display_status()  /* display simulator status */
{
	dispinst();		/* display disassembled inst */
	dispreg(regdisptop);	/* display SR */
	dispmem(memdisptop);	/* display memory */
}

error(errortype)
char	*errortype;
{
	mvprintw(ERRORLINE,0,"ERROR! -> %s ",errortype);
	dispflush();
	errorflag = TRUE;
}

dispinit()	/* initialize display */
{
	initscr();
	noecho();	/* set echo off */
}

dispflush()	/* flush display */
{
	refresh();
}

waitusec(usec)	/* wait for usec passing. */
int	usec;	/* sleep micro-sec. */
{
	usleep(usec);
}

dispfinish()	/* close display */
{
	endwin();
}

void clearline(line)	/* clear from top to end of specified line */
int	line;
{
	move(line,0);
	clrtoeol();
}

void dispreg(top)/* display registers */
int	top;	/* display start SR number */
{
	int	i;
	int	line;
#ifdef STACK_ARCH
	stack	*sptr;
#endif

	line = REGLINE + 1;

#ifdef ACCUM_ARCH
	mvprintw (REGLINE,0,"Accumlator: %04X(%6d)",
			     ACC & 0xffff,ACC); 
	clrtoeol();
#endif
#ifdef STACK_ARCH
	mvprintw (REGLINE,0,"Stack:"); clrtoeol();
        move (line,0);
	if (stacktop == NULL) {
	    printw ("stack is empty.");
	    clrtoeol();
	    line++;
	} else {
	    sptr = stacktop;
	    i = 0;
	    while (sptr != NULL) {
		printw ("S%02d:%04X(%6d) ",
			i,sptr->val & 0xffff,sptr->val);
		clrtoeol();
		if ((++i % 3) == 0) {
		    line++;
		    move(line,0);
		    clrtoeol();
		}
		sptr = sptr->next;
	    }
	    clrtoeol();
	    line++;
	}
#endif
#ifdef CISC_ARCH
	mvprintw (REGLINE,0,"Registers:"); clrtoeol();
	for(i=0 ; i<8 ; i+=2) {
	    mvprintw (line,0,"R%1d:%04X(%6d)",
			i,regfile[i] & 0xffff,regfile[i]);
	    clrtoeol();
	    mvprintw (line,40,"R%1d:%04X(%6d)",
			i+1,regfile[i+1] & 0xffff,regfile[i+1]);
	    clrtoeol();
	    line++;
	}
#endif
#ifdef RISC_ARCH
	mvprintw (REGLINE,0,"Registers:"); clrtoeol();
	for(i=top ; i<top+20 ; i+=2) {
	    mvprintw (line,0,"R%02d:%04X(%6d)",
			i,regfile[i] & 0xffff,regfile[i]);
	    clrtoeol();
	    mvprintw (line,40,"R%02d:%04X(%6d)",
			i+1,regfile[i+1] & 0xffff,regfile[i+1]);
	    clrtoeol();
	    line++;
	}
#endif
	clearline(line);
}

void dispmem(top)	/* display memory contents */
int	top;		/* display start address */
{
	int	i;
	int	line;

	line = MEMLINE + 1;

	mvprintw (MEMLINE,0,"Memories :"); clrtoeol();
	for(i=top ; i<top+20 ; i+=2) {
	    mvprintw (line,0,"Mem[%04X(%6d)]:%04X(%6d)",
			i & 0xffff,i,memory[i] & 0xffff,memory[i]);
	    clrtoeol();
	    mvprintw (line,40,"Mem[%04X(%6d)]:%04X(%6d)",
			i+1 & 0xffff,i+1,memory[i+1] & 0xffff,memory[i+1]);
	    clrtoeol();
	    line++;
	}
	clearline(line);
}

void dispinst()	/* display disassembled instruction */
{
	int	i;
	char	line[256];

	disasm(PC,line);
	mvprintw (INSTLINE,0,"Current instruction : %s",line);
	clrtoeol();

	clearline(INSTLINE+1);
}
